/*
 * linux/sound/arm/ep93xx-i2s.c -- ALSA PCM interface for the edb93xx i2s audio
 *
 * Author:      Fred Wei
 * Created:     July 19, 2005
 * Copyright:   Cirrus Logic, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>

#include <linux/version.h>
#include <linux/soundcard.h>
#include <linux/sched.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/control.h>
#include <sound/initval.h>
                                                                                                                             
#include <asm/irq.h>
#include <asm/semaphore.h>
#include <asm/hardware.h>
#include <asm/io.h>

#include <asm/arch/dma.h>
#include <asm/arch/regs_i2s.h>
#include <asm/arch/ssp.h>
#include "ep93xx-i2s.h"

//#define DEBUG 1
#ifdef DEBUG
#define DPRINTK( fmt, arg... )  printk( fmt, ##arg )
#else
#define DPRINTK( fmt, arg... )
#endif

static const snd_pcm_hardware_t ep93xx_i2s_pcm_hardware = {
	.info			= (SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_MMAP_VALID |
				   SNDRV_PCM_INFO_INTERLEAVED | SNDRV_PCM_INFO_PAUSE ),
	.formats 		= (SNDRV_PCM_FMTBIT_U8 | SNDRV_PCM_FMTBIT_S8 |
				   SNDRV_PCM_FMTBIT_U16_LE | SNDRV_PCM_FMTBIT_S16_LE |
				   SNDRV_PCM_FMTBIT_U16_BE | SNDRV_PCM_FMTBIT_S16_BE |
				   SNDRV_PCM_FMTBIT_U32_LE | SNDRV_PCM_FMTBIT_S32_LE |
				   SNDRV_PCM_FMTBIT_U32_BE | SNDRV_PCM_FMTBIT_S32_BE ),
	.rates			=( SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_11025 |
				   SNDRV_PCM_RATE_16000 | SNDRV_PCM_RATE_22050 |
				   SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 |
				   SNDRV_PCM_RATE_48000),
	.rate_min		= 8000,
	.rate_max		= 48000,
	.channels_min		= 1,
	.channels_max		= 2,					
	.period_bytes_min	= 512,
	.period_bytes_max	= 16*1024,
	.periods_min		= 2,
	.periods_max		= 8,
	.buffer_bytes_max	= 16*1024,
	.fifo_size		= 32,
};

#define AUDIO_NAME		"ep93xx-audio"

#define AUDIO_NBFRAGS_DEFAULT	8
#define AUDIO_FRAGSIZE_DEFAULT	16384

/*
 * Translates to:
 * 		s->buf_idx++;
 * 		s->buf_idx %= s->nbfrags;
 *		s->buf = s->buffers + s->buf_idx;
 *
 * So s->buf always points to the s->buf_idx-nth element of s->buffers.
 */
#define NEXT_BUF(_s_,_b_) { \
	(_s_)->_b_##_idx++; \
	(_s_)->_b_##_idx %= (_s_)->nbfrags; \
	(_s_)->_b_ = (_s_)->buffers + (_s_)->_b_##_idx; }

#define AUDIO_ACTIVE(state)	((state)->rd_ref || (state)->wr_ref)

#define AUDIO_SAMPLE_RATE_DEFAULT	44100

/*
 * OSS mixer volume levels are 16 bit with two 8 bit fields for left and right.
 * Each is in % of full.  So range is 0..100 for each of two channels.
 */ 
#define AUDIO_DEFAULT_VOLUME ( (100<<8) | 100 )

/*
 * Statics
 */
static int giOSS_Volume = AUDIO_DEFAULT_VOLUME;
static int giDAC_Volume_L = 0;
static int giDAC_Volume_R = 0;
static int giDAC_Mute = 0;

static int audio_has_been_initialized = 0;
static int audio_clocks_enabled_in_syscon = 0;
static int SSP_Handle = -1;

static int ep93xx_calc_closest_freq
(
    ulong   ulPLLFreq, 
    ulong   ulRequestedMClkFreq,
    ulong * pulActualMClkFreq,
    ulong * pulI2SDiv
);
static void ep93xx_set_samplerate(long lFrequency);
static void ep93xx_init_i2s_controller( void );

static void ep93xx_init_i2s_codec( void );
static void ep93xx_automute_i2s_codec( void );
static int ep93xx_set_volume( int val );

static void ep93xx_audio_init(void *dummy);

static audio_stream_t output_stream =
{
	.audio_num_channels 	= EP93XX_DEFAULT_NUM_CHANNELS,
	.audio_format 		= EP93XX_DEFAULT_FORMAT,
	.audio_stream_bitwidth 	= EP93XX_DEFAULT_BIT_WIDTH,
	.dma2usr_ratio 		= 2,
};

static audio_stream_t input_stream =
{
	.audio_num_channels 	= EP93XX_DEFAULT_NUM_CHANNELS,
	.audio_format 		= EP93XX_DEFAULT_FORMAT,
	.audio_stream_bitwidth 	= EP93XX_DEFAULT_BIT_WIDTH,
	.dma2usr_ratio 		= 2,
};

static audio_state_t audio_state =
{
	.output_stream		= &output_stream,
	.output_dma		= DMATx_I2S1,
	.output_id		= "I2S out",
	
	.input_stream		= &input_stream,
	.input_dma		= DMARx_I2S1,
	.input_id		= "I2S in",
	
//	.client_ioctl		= ep93xx_audio_ioctl,
	.hw_enable          	= 0,
	.hw_disable         	= 0,
	
	.set_hw_serial_format 	= 0,

//	.startup		 = ep93xx_i2s_pcm_startup,
//	.shutdown		 = ep93xx_i2s_pcm_shutdown,
//	.prepare		 = ep93xx_i2s_pcm_prepare,
	.sem			=__MUTEX_INITIALIZER(audio_state.sem),
	
	.DAC_bit_width		= 24,
	.bCompactMode		= 0,
};

static snd_pcm_t *ep93xx_i2s_pcm;

typedef struct {
    ulong   ulTotalDiv;
    ulong   ulI2SDiv;
} DIV_TABLE;

static const DIV_TABLE I2SDivTable[] =
{
    {   6, SYSCON_I2SDIV_PDIV_2  | (  2 & SYSCON_I2SDIV_MDIV_MASK) },  /* Fake entry for lower limit. */
    {   8, SYSCON_I2SDIV_PDIV_2  | (  2 & SYSCON_I2SDIV_MDIV_MASK) },
    {  10, SYSCON_I2SDIV_PDIV_25 | (  2 & SYSCON_I2SDIV_MDIV_MASK) },
    {  12, SYSCON_I2SDIV_PDIV_3  | (  2 & SYSCON_I2SDIV_MDIV_MASK) },
    {  15, SYSCON_I2SDIV_PDIV_25 | (  3 & SYSCON_I2SDIV_MDIV_MASK) },
    {  16, SYSCON_I2SDIV_PDIV_2  | (  4 & SYSCON_I2SDIV_MDIV_MASK) },
    {  18, SYSCON_I2SDIV_PDIV_3  | (  3 & SYSCON_I2SDIV_MDIV_MASK) },
    {  20, SYSCON_I2SDIV_PDIV_25 | (  4 & SYSCON_I2SDIV_MDIV_MASK) },
    {  24, SYSCON_I2SDIV_PDIV_3  | (  4 & SYSCON_I2SDIV_MDIV_MASK) },
    {  25, SYSCON_I2SDIV_PDIV_25 | (  5 & SYSCON_I2SDIV_MDIV_MASK) },
    {  28, SYSCON_I2SDIV_PDIV_2  | (  7 & SYSCON_I2SDIV_MDIV_MASK) },
    {  30, SYSCON_I2SDIV_PDIV_3  | (  5 & SYSCON_I2SDIV_MDIV_MASK) },
    {  32, SYSCON_I2SDIV_PDIV_2  | (  8 & SYSCON_I2SDIV_MDIV_MASK) },
    {  35, SYSCON_I2SDIV_PDIV_25 | (  7 & SYSCON_I2SDIV_MDIV_MASK) },
    {  36, SYSCON_I2SDIV_PDIV_3  | (  6 & SYSCON_I2SDIV_MDIV_MASK) },
    {  40, SYSCON_I2SDIV_PDIV_25 | (  8 & SYSCON_I2SDIV_MDIV_MASK) },
    {  42, SYSCON_I2SDIV_PDIV_3  | (  7 & SYSCON_I2SDIV_MDIV_MASK) },
    {  44, SYSCON_I2SDIV_PDIV_2  | ( 11 & SYSCON_I2SDIV_MDIV_MASK) },
    {  45, SYSCON_I2SDIV_PDIV_25 | (  9 & SYSCON_I2SDIV_MDIV_MASK) },
    {  48, SYSCON_I2SDIV_PDIV_3  | (  8 & SYSCON_I2SDIV_MDIV_MASK) },
    {  50, SYSCON_I2SDIV_PDIV_25 | ( 10 & SYSCON_I2SDIV_MDIV_MASK) },
    {  52, SYSCON_I2SDIV_PDIV_2  | ( 13 & SYSCON_I2SDIV_MDIV_MASK) },
    {  54, SYSCON_I2SDIV_PDIV_3  | (  9 & SYSCON_I2SDIV_MDIV_MASK) },
    {  55, SYSCON_I2SDIV_PDIV_25 | ( 11 & SYSCON_I2SDIV_MDIV_MASK) },
    {  56, SYSCON_I2SDIV_PDIV_2  | ( 14 & SYSCON_I2SDIV_MDIV_MASK) },
    {  60, SYSCON_I2SDIV_PDIV_3  | ( 10 & SYSCON_I2SDIV_MDIV_MASK) },
    {  64, SYSCON_I2SDIV_PDIV_2  | ( 16 & SYSCON_I2SDIV_MDIV_MASK) },
    {  65, SYSCON_I2SDIV_PDIV_25 | ( 13 & SYSCON_I2SDIV_MDIV_MASK) },
    {  66, SYSCON_I2SDIV_PDIV_3  | ( 11 & SYSCON_I2SDIV_MDIV_MASK) },
    {  68, SYSCON_I2SDIV_PDIV_2  | ( 17 & SYSCON_I2SDIV_MDIV_MASK) },
    {  70, SYSCON_I2SDIV_PDIV_25 | ( 14 & SYSCON_I2SDIV_MDIV_MASK) },
    {  72, SYSCON_I2SDIV_PDIV_3  | ( 12 & SYSCON_I2SDIV_MDIV_MASK) },
    {  75, SYSCON_I2SDIV_PDIV_25 | ( 15 & SYSCON_I2SDIV_MDIV_MASK) },
    {  76, SYSCON_I2SDIV_PDIV_2  | ( 19 & SYSCON_I2SDIV_MDIV_MASK) },
    {  78, SYSCON_I2SDIV_PDIV_3  | ( 13 & SYSCON_I2SDIV_MDIV_MASK) },
    {  80, SYSCON_I2SDIV_PDIV_25 | ( 16 & SYSCON_I2SDIV_MDIV_MASK) },
    {  84, SYSCON_I2SDIV_PDIV_3  | ( 14 & SYSCON_I2SDIV_MDIV_MASK) },
    {  85, SYSCON_I2SDIV_PDIV_25 | ( 17 & SYSCON_I2SDIV_MDIV_MASK) },
    {  88, SYSCON_I2SDIV_PDIV_2  | ( 22 & SYSCON_I2SDIV_MDIV_MASK) },
    {  90, SYSCON_I2SDIV_PDIV_3  | ( 15 & SYSCON_I2SDIV_MDIV_MASK) },
    {  92, SYSCON_I2SDIV_PDIV_2  | ( 23 & SYSCON_I2SDIV_MDIV_MASK) },
    {  95, SYSCON_I2SDIV_PDIV_25 | ( 19 & SYSCON_I2SDIV_MDIV_MASK) },
    {  96, SYSCON_I2SDIV_PDIV_3  | ( 16 & SYSCON_I2SDIV_MDIV_MASK) },
    { 100, SYSCON_I2SDIV_PDIV_25 | ( 20 & SYSCON_I2SDIV_MDIV_MASK) },
    { 102, SYSCON_I2SDIV_PDIV_3  | ( 17 & SYSCON_I2SDIV_MDIV_MASK) },
    { 104, SYSCON_I2SDIV_PDIV_2  | ( 26 & SYSCON_I2SDIV_MDIV_MASK) },
    { 105, SYSCON_I2SDIV_PDIV_25 | ( 21 & SYSCON_I2SDIV_MDIV_MASK) },
    { 108, SYSCON_I2SDIV_PDIV_3  | ( 18 & SYSCON_I2SDIV_MDIV_MASK) },
    { 110, SYSCON_I2SDIV_PDIV_25 | ( 22 & SYSCON_I2SDIV_MDIV_MASK) },
    { 112, SYSCON_I2SDIV_PDIV_2  | ( 28 & SYSCON_I2SDIV_MDIV_MASK) },
    { 114, SYSCON_I2SDIV_PDIV_3  | ( 19 & SYSCON_I2SDIV_MDIV_MASK) },
    { 115, SYSCON_I2SDIV_PDIV_25 | ( 23 & SYSCON_I2SDIV_MDIV_MASK) },
    { 116, SYSCON_I2SDIV_PDIV_2  | ( 29 & SYSCON_I2SDIV_MDIV_MASK) },
    { 120, SYSCON_I2SDIV_PDIV_3  | ( 20 & SYSCON_I2SDIV_MDIV_MASK) },
    { 124, SYSCON_I2SDIV_PDIV_2  | ( 31 & SYSCON_I2SDIV_MDIV_MASK) },
    { 125, SYSCON_I2SDIV_PDIV_25 | ( 25 & SYSCON_I2SDIV_MDIV_MASK) },
    { 126, SYSCON_I2SDIV_PDIV_3  | ( 21 & SYSCON_I2SDIV_MDIV_MASK) },
    { 128, SYSCON_I2SDIV_PDIV_2  | ( 32 & SYSCON_I2SDIV_MDIV_MASK) },
    { 130, SYSCON_I2SDIV_PDIV_25 | ( 26 & SYSCON_I2SDIV_MDIV_MASK) },
    { 132, SYSCON_I2SDIV_PDIV_3  | ( 22 & SYSCON_I2SDIV_MDIV_MASK) },
    { 135, SYSCON_I2SDIV_PDIV_25 | ( 27 & SYSCON_I2SDIV_MDIV_MASK) },
    { 136, SYSCON_I2SDIV_PDIV_2  | ( 34 & SYSCON_I2SDIV_MDIV_MASK) },
    { 138, SYSCON_I2SDIV_PDIV_3  | ( 23 & SYSCON_I2SDIV_MDIV_MASK) },
    { 140, SYSCON_I2SDIV_PDIV_25 | ( 28 & SYSCON_I2SDIV_MDIV_MASK) },
    { 144, SYSCON_I2SDIV_PDIV_3  | ( 24 & SYSCON_I2SDIV_MDIV_MASK) },
    { 145, SYSCON_I2SDIV_PDIV_25 | ( 29 & SYSCON_I2SDIV_MDIV_MASK) },
    { 148, SYSCON_I2SDIV_PDIV_2  | ( 37 & SYSCON_I2SDIV_MDIV_MASK) },
    { 150, SYSCON_I2SDIV_PDIV_3  | ( 25 & SYSCON_I2SDIV_MDIV_MASK) },
    { 152, SYSCON_I2SDIV_PDIV_2  | ( 38 & SYSCON_I2SDIV_MDIV_MASK) },
    { 155, SYSCON_I2SDIV_PDIV_25 | ( 31 & SYSCON_I2SDIV_MDIV_MASK) },
    { 156, SYSCON_I2SDIV_PDIV_3  | ( 26 & SYSCON_I2SDIV_MDIV_MASK) },
    { 160, SYSCON_I2SDIV_PDIV_25 | ( 32 & SYSCON_I2SDIV_MDIV_MASK) },
    { 162, SYSCON_I2SDIV_PDIV_3  | ( 27 & SYSCON_I2SDIV_MDIV_MASK) },
    { 164, SYSCON_I2SDIV_PDIV_2  | ( 41 & SYSCON_I2SDIV_MDIV_MASK) },
    { 165, SYSCON_I2SDIV_PDIV_25 | ( 33 & SYSCON_I2SDIV_MDIV_MASK) },
    { 168, SYSCON_I2SDIV_PDIV_3  | ( 28 & SYSCON_I2SDIV_MDIV_MASK) },
    { 170, SYSCON_I2SDIV_PDIV_25 | ( 34 & SYSCON_I2SDIV_MDIV_MASK) },
    { 172, SYSCON_I2SDIV_PDIV_2  | ( 43 & SYSCON_I2SDIV_MDIV_MASK) },
    { 174, SYSCON_I2SDIV_PDIV_3  | ( 29 & SYSCON_I2SDIV_MDIV_MASK) },
    { 175, SYSCON_I2SDIV_PDIV_25 | ( 35 & SYSCON_I2SDIV_MDIV_MASK) },
    { 176, SYSCON_I2SDIV_PDIV_2  | ( 44 & SYSCON_I2SDIV_MDIV_MASK) },
    { 180, SYSCON_I2SDIV_PDIV_3  | ( 30 & SYSCON_I2SDIV_MDIV_MASK) },
    { 184, SYSCON_I2SDIV_PDIV_2  | ( 46 & SYSCON_I2SDIV_MDIV_MASK) },
    { 185, SYSCON_I2SDIV_PDIV_25 | ( 37 & SYSCON_I2SDIV_MDIV_MASK) },
    { 186, SYSCON_I2SDIV_PDIV_3  | ( 31 & SYSCON_I2SDIV_MDIV_MASK) },
    { 188, SYSCON_I2SDIV_PDIV_2  | ( 47 & SYSCON_I2SDIV_MDIV_MASK) },
    { 190, SYSCON_I2SDIV_PDIV_25 | ( 38 & SYSCON_I2SDIV_MDIV_MASK) },
    { 192, SYSCON_I2SDIV_PDIV_3  | ( 32 & SYSCON_I2SDIV_MDIV_MASK) },
    { 195, SYSCON_I2SDIV_PDIV_25 | ( 39 & SYSCON_I2SDIV_MDIV_MASK) },
    { 196, SYSCON_I2SDIV_PDIV_2  | ( 49 & SYSCON_I2SDIV_MDIV_MASK) },
    { 198, SYSCON_I2SDIV_PDIV_3  | ( 33 & SYSCON_I2SDIV_MDIV_MASK) },
    { 200, SYSCON_I2SDIV_PDIV_25 | ( 40 & SYSCON_I2SDIV_MDIV_MASK) },
    { 204, SYSCON_I2SDIV_PDIV_3  | ( 34 & SYSCON_I2SDIV_MDIV_MASK) },
    { 205, SYSCON_I2SDIV_PDIV_25 | ( 41 & SYSCON_I2SDIV_MDIV_MASK) },
    { 208, SYSCON_I2SDIV_PDIV_2  | ( 52 & SYSCON_I2SDIV_MDIV_MASK) },
    { 210, SYSCON_I2SDIV_PDIV_3  | ( 35 & SYSCON_I2SDIV_MDIV_MASK) },
    { 212, SYSCON_I2SDIV_PDIV_2  | ( 53 & SYSCON_I2SDIV_MDIV_MASK) },
    { 215, SYSCON_I2SDIV_PDIV_25 | ( 43 & SYSCON_I2SDIV_MDIV_MASK) },
    { 216, SYSCON_I2SDIV_PDIV_3  | ( 36 & SYSCON_I2SDIV_MDIV_MASK) },
    { 220, SYSCON_I2SDIV_PDIV_25 | ( 44 & SYSCON_I2SDIV_MDIV_MASK) },
    { 222, SYSCON_I2SDIV_PDIV_3  | ( 37 & SYSCON_I2SDIV_MDIV_MASK) },
    { 224, SYSCON_I2SDIV_PDIV_2  | ( 56 & SYSCON_I2SDIV_MDIV_MASK) },
    { 225, SYSCON_I2SDIV_PDIV_25 | ( 45 & SYSCON_I2SDIV_MDIV_MASK) },
    { 228, SYSCON_I2SDIV_PDIV_3  | ( 38 & SYSCON_I2SDIV_MDIV_MASK) },
    { 230, SYSCON_I2SDIV_PDIV_25 | ( 46 & SYSCON_I2SDIV_MDIV_MASK) },
    { 232, SYSCON_I2SDIV_PDIV_2  | ( 58 & SYSCON_I2SDIV_MDIV_MASK) },
    { 234, SYSCON_I2SDIV_PDIV_3  | ( 39 & SYSCON_I2SDIV_MDIV_MASK) },
    { 235, SYSCON_I2SDIV_PDIV_25 | ( 47 & SYSCON_I2SDIV_MDIV_MASK) },
    { 236, SYSCON_I2SDIV_PDIV_2  | ( 59 & SYSCON_I2SDIV_MDIV_MASK) },
    { 240, SYSCON_I2SDIV_PDIV_3  | ( 40 & SYSCON_I2SDIV_MDIV_MASK) },
    { 244, SYSCON_I2SDIV_PDIV_2  | ( 61 & SYSCON_I2SDIV_MDIV_MASK) },
    { 245, SYSCON_I2SDIV_PDIV_25 | ( 49 & SYSCON_I2SDIV_MDIV_MASK) },
    { 246, SYSCON_I2SDIV_PDIV_3  | ( 41 & SYSCON_I2SDIV_MDIV_MASK) },
    { 248, SYSCON_I2SDIV_PDIV_2  | ( 62 & SYSCON_I2SDIV_MDIV_MASK) },
    { 250, SYSCON_I2SDIV_PDIV_25 | ( 50 & SYSCON_I2SDIV_MDIV_MASK) },
    { 252, SYSCON_I2SDIV_PDIV_3  | ( 42 & SYSCON_I2SDIV_MDIV_MASK) },
    { 255, SYSCON_I2SDIV_PDIV_25 | ( 51 & SYSCON_I2SDIV_MDIV_MASK) },
    { 256, SYSCON_I2SDIV_PDIV_2  | ( 64 & SYSCON_I2SDIV_MDIV_MASK) },
    { 258, SYSCON_I2SDIV_PDIV_3  | ( 43 & SYSCON_I2SDIV_MDIV_MASK) },
    { 260, SYSCON_I2SDIV_PDIV_25 | ( 52 & SYSCON_I2SDIV_MDIV_MASK) },
    { 264, SYSCON_I2SDIV_PDIV_3  | ( 44 & SYSCON_I2SDIV_MDIV_MASK) },
    { 265, SYSCON_I2SDIV_PDIV_25 | ( 53 & SYSCON_I2SDIV_MDIV_MASK) },
    { 268, SYSCON_I2SDIV_PDIV_2  | ( 67 & SYSCON_I2SDIV_MDIV_MASK) },
    { 270, SYSCON_I2SDIV_PDIV_3  | ( 45 & SYSCON_I2SDIV_MDIV_MASK) },
    { 272, SYSCON_I2SDIV_PDIV_2  | ( 68 & SYSCON_I2SDIV_MDIV_MASK) },
    { 275, SYSCON_I2SDIV_PDIV_25 | ( 55 & SYSCON_I2SDIV_MDIV_MASK) },
    { 276, SYSCON_I2SDIV_PDIV_3  | ( 46 & SYSCON_I2SDIV_MDIV_MASK) },
    { 280, SYSCON_I2SDIV_PDIV_25 | ( 56 & SYSCON_I2SDIV_MDIV_MASK) },
    { 282, SYSCON_I2SDIV_PDIV_3  | ( 47 & SYSCON_I2SDIV_MDIV_MASK) },
    { 284, SYSCON_I2SDIV_PDIV_2  | ( 71 & SYSCON_I2SDIV_MDIV_MASK) },
    { 285, SYSCON_I2SDIV_PDIV_25 | ( 57 & SYSCON_I2SDIV_MDIV_MASK) },
    { 288, SYSCON_I2SDIV_PDIV_3  | ( 48 & SYSCON_I2SDIV_MDIV_MASK) },
    { 290, SYSCON_I2SDIV_PDIV_25 | ( 58 & SYSCON_I2SDIV_MDIV_MASK) },
    { 292, SYSCON_I2SDIV_PDIV_2  | ( 73 & SYSCON_I2SDIV_MDIV_MASK) },
    { 294, SYSCON_I2SDIV_PDIV_3  | ( 49 & SYSCON_I2SDIV_MDIV_MASK) },
    { 295, SYSCON_I2SDIV_PDIV_25 | ( 59 & SYSCON_I2SDIV_MDIV_MASK) },
    { 296, SYSCON_I2SDIV_PDIV_2  | ( 74 & SYSCON_I2SDIV_MDIV_MASK) },
    { 300, SYSCON_I2SDIV_PDIV_3  | ( 50 & SYSCON_I2SDIV_MDIV_MASK) },
    { 304, SYSCON_I2SDIV_PDIV_2  | ( 76 & SYSCON_I2SDIV_MDIV_MASK) },
    { 305, SYSCON_I2SDIV_PDIV_25 | ( 61 & SYSCON_I2SDIV_MDIV_MASK) },
    { 306, SYSCON_I2SDIV_PDIV_3  | ( 51 & SYSCON_I2SDIV_MDIV_MASK) },
    { 308, SYSCON_I2SDIV_PDIV_2  | ( 77 & SYSCON_I2SDIV_MDIV_MASK) },
    { 310, SYSCON_I2SDIV_PDIV_25 | ( 62 & SYSCON_I2SDIV_MDIV_MASK) },
    { 312, SYSCON_I2SDIV_PDIV_3  | ( 52 & SYSCON_I2SDIV_MDIV_MASK) },
    { 315, SYSCON_I2SDIV_PDIV_25 | ( 63 & SYSCON_I2SDIV_MDIV_MASK) },
    { 316, SYSCON_I2SDIV_PDIV_2  | ( 79 & SYSCON_I2SDIV_MDIV_MASK) },
    { 318, SYSCON_I2SDIV_PDIV_3  | ( 53 & SYSCON_I2SDIV_MDIV_MASK) },
    { 320, SYSCON_I2SDIV_PDIV_25 | ( 64 & SYSCON_I2SDIV_MDIV_MASK) },
    { 324, SYSCON_I2SDIV_PDIV_3  | ( 54 & SYSCON_I2SDIV_MDIV_MASK) },
    { 325, SYSCON_I2SDIV_PDIV_25 | ( 65 & SYSCON_I2SDIV_MDIV_MASK) },
    { 328, SYSCON_I2SDIV_PDIV_2  | ( 82 & SYSCON_I2SDIV_MDIV_MASK) },
    { 330, SYSCON_I2SDIV_PDIV_3  | ( 55 & SYSCON_I2SDIV_MDIV_MASK) },
    { 332, SYSCON_I2SDIV_PDIV_2  | ( 83 & SYSCON_I2SDIV_MDIV_MASK) },
    { 335, SYSCON_I2SDIV_PDIV_25 | ( 67 & SYSCON_I2SDIV_MDIV_MASK) },
    { 336, SYSCON_I2SDIV_PDIV_3  | ( 56 & SYSCON_I2SDIV_MDIV_MASK) },
    { 340, SYSCON_I2SDIV_PDIV_25 | ( 68 & SYSCON_I2SDIV_MDIV_MASK) },
    { 342, SYSCON_I2SDIV_PDIV_3  | ( 57 & SYSCON_I2SDIV_MDIV_MASK) },
    { 344, SYSCON_I2SDIV_PDIV_2  | ( 86 & SYSCON_I2SDIV_MDIV_MASK) },
    { 345, SYSCON_I2SDIV_PDIV_25 | ( 69 & SYSCON_I2SDIV_MDIV_MASK) },
    { 348, SYSCON_I2SDIV_PDIV_3  | ( 58 & SYSCON_I2SDIV_MDIV_MASK) },
    { 350, SYSCON_I2SDIV_PDIV_25 | ( 70 & SYSCON_I2SDIV_MDIV_MASK) },
    { 352, SYSCON_I2SDIV_PDIV_2  | ( 88 & SYSCON_I2SDIV_MDIV_MASK) },
    { 354, SYSCON_I2SDIV_PDIV_3  | ( 59 & SYSCON_I2SDIV_MDIV_MASK) },
    { 355, SYSCON_I2SDIV_PDIV_25 | ( 71 & SYSCON_I2SDIV_MDIV_MASK) },
    { 356, SYSCON_I2SDIV_PDIV_2  | ( 89 & SYSCON_I2SDIV_MDIV_MASK) },
    { 360, SYSCON_I2SDIV_PDIV_3  | ( 60 & SYSCON_I2SDIV_MDIV_MASK) },
    { 364, SYSCON_I2SDIV_PDIV_2  | ( 91 & SYSCON_I2SDIV_MDIV_MASK) },
    { 365, SYSCON_I2SDIV_PDIV_25 | ( 73 & SYSCON_I2SDIV_MDIV_MASK) },
    { 366, SYSCON_I2SDIV_PDIV_3  | ( 61 & SYSCON_I2SDIV_MDIV_MASK) },
    { 368, SYSCON_I2SDIV_PDIV_2  | ( 92 & SYSCON_I2SDIV_MDIV_MASK) },
    { 370, SYSCON_I2SDIV_PDIV_25 | ( 74 & SYSCON_I2SDIV_MDIV_MASK) },
    { 372, SYSCON_I2SDIV_PDIV_3  | ( 62 & SYSCON_I2SDIV_MDIV_MASK) },
    { 375, SYSCON_I2SDIV_PDIV_25 | ( 75 & SYSCON_I2SDIV_MDIV_MASK) },
    { 376, SYSCON_I2SDIV_PDIV_2  | ( 94 & SYSCON_I2SDIV_MDIV_MASK) },
    { 378, SYSCON_I2SDIV_PDIV_3  | ( 63 & SYSCON_I2SDIV_MDIV_MASK) },
    { 380, SYSCON_I2SDIV_PDIV_25 | ( 76 & SYSCON_I2SDIV_MDIV_MASK) },
    { 384, SYSCON_I2SDIV_PDIV_3  | ( 64 & SYSCON_I2SDIV_MDIV_MASK) },
    { 385, SYSCON_I2SDIV_PDIV_25 | ( 77 & SYSCON_I2SDIV_MDIV_MASK) },
    { 388, SYSCON_I2SDIV_PDIV_2  | ( 97 & SYSCON_I2SDIV_MDIV_MASK) },
    { 390, SYSCON_I2SDIV_PDIV_3  | ( 65 & SYSCON_I2SDIV_MDIV_MASK) },
    { 392, SYSCON_I2SDIV_PDIV_2  | ( 98 & SYSCON_I2SDIV_MDIV_MASK) },
    { 395, SYSCON_I2SDIV_PDIV_25 | ( 79 & SYSCON_I2SDIV_MDIV_MASK) },
    { 396, SYSCON_I2SDIV_PDIV_3  | ( 66 & SYSCON_I2SDIV_MDIV_MASK) },
    { 400, SYSCON_I2SDIV_PDIV_25 | ( 80 & SYSCON_I2SDIV_MDIV_MASK) },
    { 402, SYSCON_I2SDIV_PDIV_3  | ( 67 & SYSCON_I2SDIV_MDIV_MASK) },
    { 404, SYSCON_I2SDIV_PDIV_2  | (101 & SYSCON_I2SDIV_MDIV_MASK) },
    { 405, SYSCON_I2SDIV_PDIV_25 | ( 81 & SYSCON_I2SDIV_MDIV_MASK) },
    { 408, SYSCON_I2SDIV_PDIV_3  | ( 68 & SYSCON_I2SDIV_MDIV_MASK) },
    { 410, SYSCON_I2SDIV_PDIV_25 | ( 82 & SYSCON_I2SDIV_MDIV_MASK) },
    { 412, SYSCON_I2SDIV_PDIV_2  | (103 & SYSCON_I2SDIV_MDIV_MASK) },
    { 414, SYSCON_I2SDIV_PDIV_3  | ( 69 & SYSCON_I2SDIV_MDIV_MASK) },
    { 415, SYSCON_I2SDIV_PDIV_25 | ( 83 & SYSCON_I2SDIV_MDIV_MASK) },
    { 416, SYSCON_I2SDIV_PDIV_2  | (104 & SYSCON_I2SDIV_MDIV_MASK) },
    { 420, SYSCON_I2SDIV_PDIV_3  | ( 70 & SYSCON_I2SDIV_MDIV_MASK) },
    { 424, SYSCON_I2SDIV_PDIV_2  | (106 & SYSCON_I2SDIV_MDIV_MASK) },
    { 425, SYSCON_I2SDIV_PDIV_25 | ( 85 & SYSCON_I2SDIV_MDIV_MASK) },
    { 426, SYSCON_I2SDIV_PDIV_3  | ( 71 & SYSCON_I2SDIV_MDIV_MASK) },
    { 428, SYSCON_I2SDIV_PDIV_2  | (107 & SYSCON_I2SDIV_MDIV_MASK) },
    { 430, SYSCON_I2SDIV_PDIV_25 | ( 86 & SYSCON_I2SDIV_MDIV_MASK) },
    { 432, SYSCON_I2SDIV_PDIV_3  | ( 72 & SYSCON_I2SDIV_MDIV_MASK) },
    { 435, SYSCON_I2SDIV_PDIV_25 | ( 87 & SYSCON_I2SDIV_MDIV_MASK) },
    { 436, SYSCON_I2SDIV_PDIV_2  | (109 & SYSCON_I2SDIV_MDIV_MASK) },
    { 438, SYSCON_I2SDIV_PDIV_3  | ( 73 & SYSCON_I2SDIV_MDIV_MASK) },
    { 440, SYSCON_I2SDIV_PDIV_25 | ( 88 & SYSCON_I2SDIV_MDIV_MASK) },
    { 444, SYSCON_I2SDIV_PDIV_3  | ( 74 & SYSCON_I2SDIV_MDIV_MASK) },
    { 445, SYSCON_I2SDIV_PDIV_25 | ( 89 & SYSCON_I2SDIV_MDIV_MASK) },
    { 448, SYSCON_I2SDIV_PDIV_2  | (112 & SYSCON_I2SDIV_MDIV_MASK) },
    { 450, SYSCON_I2SDIV_PDIV_3  | ( 75 & SYSCON_I2SDIV_MDIV_MASK) },
    { 452, SYSCON_I2SDIV_PDIV_2  | (113 & SYSCON_I2SDIV_MDIV_MASK) },
    { 455, SYSCON_I2SDIV_PDIV_25 | ( 91 & SYSCON_I2SDIV_MDIV_MASK) },
    { 456, SYSCON_I2SDIV_PDIV_3  | ( 76 & SYSCON_I2SDIV_MDIV_MASK) },
    { 460, SYSCON_I2SDIV_PDIV_25 | ( 92 & SYSCON_I2SDIV_MDIV_MASK) },
    { 462, SYSCON_I2SDIV_PDIV_3  | ( 77 & SYSCON_I2SDIV_MDIV_MASK) },
    { 464, SYSCON_I2SDIV_PDIV_2  | (116 & SYSCON_I2SDIV_MDIV_MASK) },
    { 465, SYSCON_I2SDIV_PDIV_25 | ( 93 & SYSCON_I2SDIV_MDIV_MASK) },
    { 468, SYSCON_I2SDIV_PDIV_3  | ( 78 & SYSCON_I2SDIV_MDIV_MASK) },
    { 470, SYSCON_I2SDIV_PDIV_25 | ( 94 & SYSCON_I2SDIV_MDIV_MASK) },
    { 472, SYSCON_I2SDIV_PDIV_2  | (118 & SYSCON_I2SDIV_MDIV_MASK) },
    { 474, SYSCON_I2SDIV_PDIV_3  | ( 79 & SYSCON_I2SDIV_MDIV_MASK) },
    { 475, SYSCON_I2SDIV_PDIV_25 | ( 95 & SYSCON_I2SDIV_MDIV_MASK) },
    { 476, SYSCON_I2SDIV_PDIV_2  | (119 & SYSCON_I2SDIV_MDIV_MASK) },
    { 480, SYSCON_I2SDIV_PDIV_3  | ( 80 & SYSCON_I2SDIV_MDIV_MASK) },
    { 484, SYSCON_I2SDIV_PDIV_2  | (121 & SYSCON_I2SDIV_MDIV_MASK) },
    { 485, SYSCON_I2SDIV_PDIV_25 | ( 97 & SYSCON_I2SDIV_MDIV_MASK) },
    { 486, SYSCON_I2SDIV_PDIV_3  | ( 81 & SYSCON_I2SDIV_MDIV_MASK) },
    { 488, SYSCON_I2SDIV_PDIV_2  | (122 & SYSCON_I2SDIV_MDIV_MASK) },
    { 490, SYSCON_I2SDIV_PDIV_25 | ( 98 & SYSCON_I2SDIV_MDIV_MASK) },
    { 492, SYSCON_I2SDIV_PDIV_3  | ( 82 & SYSCON_I2SDIV_MDIV_MASK) },
    { 495, SYSCON_I2SDIV_PDIV_25 | ( 99 & SYSCON_I2SDIV_MDIV_MASK) },
    { 496, SYSCON_I2SDIV_PDIV_2  | (124 & SYSCON_I2SDIV_MDIV_MASK) },
    { 498, SYSCON_I2SDIV_PDIV_3  | ( 83 & SYSCON_I2SDIV_MDIV_MASK) },
    { 500, SYSCON_I2SDIV_PDIV_25 | (100 & SYSCON_I2SDIV_MDIV_MASK) },
    { 504, SYSCON_I2SDIV_PDIV_3  | ( 84 & SYSCON_I2SDIV_MDIV_MASK) },
    { 505, SYSCON_I2SDIV_PDIV_25 | (101 & SYSCON_I2SDIV_MDIV_MASK) },
    { 508, SYSCON_I2SDIV_PDIV_2  | (127 & SYSCON_I2SDIV_MDIV_MASK) },
    { 510, SYSCON_I2SDIV_PDIV_3  | ( 85 & SYSCON_I2SDIV_MDIV_MASK) },
    { 515, SYSCON_I2SDIV_PDIV_25 | (103 & SYSCON_I2SDIV_MDIV_MASK) },
    { 516, SYSCON_I2SDIV_PDIV_3  | ( 86 & SYSCON_I2SDIV_MDIV_MASK) },
    { 520, SYSCON_I2SDIV_PDIV_25 | (104 & SYSCON_I2SDIV_MDIV_MASK) },
    { 522, SYSCON_I2SDIV_PDIV_3  | ( 87 & SYSCON_I2SDIV_MDIV_MASK) },
    { 525, SYSCON_I2SDIV_PDIV_25 | (105 & SYSCON_I2SDIV_MDIV_MASK) },
    { 528, SYSCON_I2SDIV_PDIV_3  | ( 88 & SYSCON_I2SDIV_MDIV_MASK) },
    { 530, SYSCON_I2SDIV_PDIV_25 | (106 & SYSCON_I2SDIV_MDIV_MASK) },
    { 534, SYSCON_I2SDIV_PDIV_3  | ( 89 & SYSCON_I2SDIV_MDIV_MASK) },
    { 535, SYSCON_I2SDIV_PDIV_25 | (107 & SYSCON_I2SDIV_MDIV_MASK) },
    { 540, SYSCON_I2SDIV_PDIV_3  | ( 90 & SYSCON_I2SDIV_MDIV_MASK) },
    { 545, SYSCON_I2SDIV_PDIV_25 | (109 & SYSCON_I2SDIV_MDIV_MASK) },
    { 546, SYSCON_I2SDIV_PDIV_3  | ( 91 & SYSCON_I2SDIV_MDIV_MASK) },
    { 550, SYSCON_I2SDIV_PDIV_25 | (110 & SYSCON_I2SDIV_MDIV_MASK) },
    { 552, SYSCON_I2SDIV_PDIV_3  | ( 92 & SYSCON_I2SDIV_MDIV_MASK) },
    { 555, SYSCON_I2SDIV_PDIV_25 | (111 & SYSCON_I2SDIV_MDIV_MASK) },
    { 558, SYSCON_I2SDIV_PDIV_3  | ( 93 & SYSCON_I2SDIV_MDIV_MASK) },
    { 560, SYSCON_I2SDIV_PDIV_25 | (112 & SYSCON_I2SDIV_MDIV_MASK) },
    { 564, SYSCON_I2SDIV_PDIV_3  | ( 94 & SYSCON_I2SDIV_MDIV_MASK) },
    { 565, SYSCON_I2SDIV_PDIV_25 | (113 & SYSCON_I2SDIV_MDIV_MASK) },
    { 570, SYSCON_I2SDIV_PDIV_3  | ( 95 & SYSCON_I2SDIV_MDIV_MASK) },
    { 575, SYSCON_I2SDIV_PDIV_25 | (115 & SYSCON_I2SDIV_MDIV_MASK) },
    { 576, SYSCON_I2SDIV_PDIV_3  | ( 96 & SYSCON_I2SDIV_MDIV_MASK) },
    { 580, SYSCON_I2SDIV_PDIV_25 | (116 & SYSCON_I2SDIV_MDIV_MASK) },
    { 582, SYSCON_I2SDIV_PDIV_3  | ( 97 & SYSCON_I2SDIV_MDIV_MASK) },
    { 585, SYSCON_I2SDIV_PDIV_25 | (117 & SYSCON_I2SDIV_MDIV_MASK) },
    { 588, SYSCON_I2SDIV_PDIV_3  | ( 98 & SYSCON_I2SDIV_MDIV_MASK) },
    { 590, SYSCON_I2SDIV_PDIV_25 | (118 & SYSCON_I2SDIV_MDIV_MASK) },
    { 594, SYSCON_I2SDIV_PDIV_3  | ( 99 & SYSCON_I2SDIV_MDIV_MASK) },
    { 595, SYSCON_I2SDIV_PDIV_25 | (119 & SYSCON_I2SDIV_MDIV_MASK) },
    { 600, SYSCON_I2SDIV_PDIV_3  | (100 & SYSCON_I2SDIV_MDIV_MASK) },
    { 605, SYSCON_I2SDIV_PDIV_25 | (121 & SYSCON_I2SDIV_MDIV_MASK) },
    { 606, SYSCON_I2SDIV_PDIV_3  | (101 & SYSCON_I2SDIV_MDIV_MASK) },
    { 610, SYSCON_I2SDIV_PDIV_25 | (122 & SYSCON_I2SDIV_MDIV_MASK) },
    { 612, SYSCON_I2SDIV_PDIV_3  | (102 & SYSCON_I2SDIV_MDIV_MASK) },
    { 615, SYSCON_I2SDIV_PDIV_25 | (123 & SYSCON_I2SDIV_MDIV_MASK) },
    { 618, SYSCON_I2SDIV_PDIV_3  | (103 & SYSCON_I2SDIV_MDIV_MASK) },
    { 620, SYSCON_I2SDIV_PDIV_25 | (124 & SYSCON_I2SDIV_MDIV_MASK) },
    { 624, SYSCON_I2SDIV_PDIV_3  | (104 & SYSCON_I2SDIV_MDIV_MASK) },
    { 625, SYSCON_I2SDIV_PDIV_25 | (125 & SYSCON_I2SDIV_MDIV_MASK) },
    { 630, SYSCON_I2SDIV_PDIV_3  | (105 & SYSCON_I2SDIV_MDIV_MASK) },
    { 635, SYSCON_I2SDIV_PDIV_25 | (127 & SYSCON_I2SDIV_MDIV_MASK) },
    { 636, SYSCON_I2SDIV_PDIV_3  | (106 & SYSCON_I2SDIV_MDIV_MASK) },
    { 642, SYSCON_I2SDIV_PDIV_3  | (107 & SYSCON_I2SDIV_MDIV_MASK) },
    { 648, SYSCON_I2SDIV_PDIV_3  | (108 & SYSCON_I2SDIV_MDIV_MASK) },
    { 654, SYSCON_I2SDIV_PDIV_3  | (109 & SYSCON_I2SDIV_MDIV_MASK) },
    { 660, SYSCON_I2SDIV_PDIV_3  | (110 & SYSCON_I2SDIV_MDIV_MASK) },
    { 666, SYSCON_I2SDIV_PDIV_3  | (111 & SYSCON_I2SDIV_MDIV_MASK) },
    { 672, SYSCON_I2SDIV_PDIV_3  | (112 & SYSCON_I2SDIV_MDIV_MASK) },
    { 678, SYSCON_I2SDIV_PDIV_3  | (113 & SYSCON_I2SDIV_MDIV_MASK) },
    { 684, SYSCON_I2SDIV_PDIV_3  | (114 & SYSCON_I2SDIV_MDIV_MASK) },
    { 690, SYSCON_I2SDIV_PDIV_3  | (115 & SYSCON_I2SDIV_MDIV_MASK) },
    { 696, SYSCON_I2SDIV_PDIV_3  | (116 & SYSCON_I2SDIV_MDIV_MASK) },
    { 702, SYSCON_I2SDIV_PDIV_3  | (117 & SYSCON_I2SDIV_MDIV_MASK) },
    { 708, SYSCON_I2SDIV_PDIV_3  | (118 & SYSCON_I2SDIV_MDIV_MASK) },
    { 714, SYSCON_I2SDIV_PDIV_3  | (119 & SYSCON_I2SDIV_MDIV_MASK) },
    { 720, SYSCON_I2SDIV_PDIV_3  | (120 & SYSCON_I2SDIV_MDIV_MASK) },
    { 726, SYSCON_I2SDIV_PDIV_3  | (121 & SYSCON_I2SDIV_MDIV_MASK) },
    { 732, SYSCON_I2SDIV_PDIV_3  | (122 & SYSCON_I2SDIV_MDIV_MASK) },
    { 738, SYSCON_I2SDIV_PDIV_3  | (123 & SYSCON_I2SDIV_MDIV_MASK) },
    { 744, SYSCON_I2SDIV_PDIV_3  | (124 & SYSCON_I2SDIV_MDIV_MASK) },
    { 750, SYSCON_I2SDIV_PDIV_3  | (125 & SYSCON_I2SDIV_MDIV_MASK) },
    { 756, SYSCON_I2SDIV_PDIV_3  | (126 & SYSCON_I2SDIV_MDIV_MASK) },
    { 762, SYSCON_I2SDIV_PDIV_3  | (127 & SYSCON_I2SDIV_MDIV_MASK) }
};

/*
 * When we get to the multichannel case the pre-fill and enable code
 * will go to the dma driver's start routine.
 */
static void ep93xx_audio_enable(void)
{
	int i;
	unsigned long ulTemp;
	
	DPRINTK("ep93xx_audio_enable - enter\n");

	/*
	 * Disable the channels, disable the i2s controller to clear out the 
	 * fifo.  Disabling the controller will clear the flags	in i2s_csr.
	 */
	outl( 0, I2STX0En );
	outl( 0, I2STX1En );
	outl( 0, I2STX2En );
	outl( 0, I2SRX0En );
	outl( 0, I2SRX1En );
	outl( 0, I2SRX2En );
	ulTemp = inl( I2SRX2En ); /* input to push the outl's past the wrapper */
	
	/*
	 * Disable the i2s controller to clear out the fifo and clear
	 * the flags in i2s_csr.
	 */
	outl( 0, I2SGlCtrl );
	ulTemp = inl( I2SGlCtrl );

	/*
	 * Enable the i2s controller.  It is important not to just
	 * leave the i2s controller enabled and skip this step. 
	 */
	outl( 1, I2SGlCtrl );
	ulTemp = inl( I2SGlCtrl );

	/*
	 * Prefill the tx fifo with zeros.
	 */
	for( i = 0 ; i < 8 ; i++ )
	{
		outl( 0, I2STX0Lft );
		outl( 0, I2STX0Rt );
		outl( 0, I2STX1Lft );
		outl( 0, I2STX1Rt );
		outl( 0, I2STX2Lft );
		outl( 0, I2STX2Rt );
		ulTemp = inl( I2SRX2En ); /* input to push the outl's past the wrapper */
	}

	/*
	 * Enable the tx channels.
	 */
	outl( 1, I2STX0En );
	outl( 1, I2STX1En );
	outl( 1, I2STX2En );
	outl( 1, I2SRX0En );
	outl( 1, I2SRX1En );
	outl( 1, I2SRX2En );
	ulTemp = inl( I2SRX2En ); /* input to push the outl's past the wrapper */

	DPRINTK("ep93xx_audio_enable - EXIT\n");

}

static void ep93xx_audio_disable(void *dummy)
{
	DPRINTK("ep93xx_audio_disable - enter\n");

	audio_has_been_initialized = 0;

	/*
	 * Disable the channels, disable the i2s controller to clear out the 
	 * fifo.  Disabling the controller will clear the flags	in i2s_csr.
	 */
	outl( 0, I2STX0En );
	outl( 0, I2STX1En );
	outl( 0, I2STX2En );
	outl( 0, I2SRX0En );
	outl( 0, I2SRX1En );
	outl( 0, I2SRX2En );
	outl( 0, I2SGlCtrl );

	DPRINTK("ep93xx_audio_disable - EXIT\n");
}


/*
 * ep93xx_calc_closest_freq
 * 
 *   ulPLLFreq           -  PLL output Frequency (Mhz)
 *   ulRequestedMClkFreq -  Requested Video Clock Frequency.
 *   pulActualMClkFreq   -  Returned Actual Video Rate.
 *   pulI2SDiv           -  Video Divider register.
 *
 *   Return            0 - Failure
 *                     1 - Success
 */
static int ep93xx_calc_closest_freq
(
    ulong   ulPLLFreq, 
    ulong   ulRequestedMClkFreq,
    ulong * pulActualMClkFreq,
    ulong * pulI2SDiv
)
{
	ulong   ulLower;
	ulong   ulUpper;
	ulong   ulDiv;
	int     x;

	/*
	 * Calculate the closest divisor.
	 */
	ulDiv =  (ulPLLFreq * 2)/ ulRequestedMClkFreq;

	for(x = 1; x < sizeof(I2SDivTable)/sizeof(DIV_TABLE); x++)
	{

		/*
		 * Calculate the next greater and lower value.
		 */
		ulLower = I2SDivTable[x - 1].ulTotalDiv;     
		ulUpper = I2SDivTable[x].ulTotalDiv;     

		/*
		 * Check to see if it is in between the two values.
		 */
		if(ulLower <= ulDiv && ulDiv < ulUpper)
		{
			break;
		}
	}

	/*
	 * Return if we did not find a divisor.
	 */
	if(x == sizeof(I2SDivTable)/sizeof(DIV_TABLE))
	{
		DPRINTK("couldn't find a divisor.\n");

		*pulActualMClkFreq  = 0;
		*pulI2SDiv          = 0;
		return -1;
	}

	/*
	 * See which is closer, the upper or the lower case.
	 */
	if(ulUpper * ulRequestedMClkFreq - ulPLLFreq * 2 >  
		ulPLLFreq * 2 - ulLower * ulRequestedMClkFreq)
	{
		x-=1;
	}
	*pulActualMClkFreq  = (ulPLLFreq * 2)/ I2SDivTable[x].ulTotalDiv;
	*pulI2SDiv          = I2SDivTable[x].ulI2SDiv;

	return 0;
}

/*
 * ep93xx_get_PLL_frequency
 *
 * Given a value for ClkSet1 or ClkSet2, calculate the PLL1 or PLL2 frequency.
 */
static ulong ep93xx_get_PLL_frequency( ulong ulCLKSET )
{
	ulong ulX1FBD, ulX2FBD, ulX2IPD, ulPS, ulPLL_Freq;

	ulPS = (ulCLKSET & SYSCON_CLKSET1_PLL1_PS_MASK) >> SYSCON_CLKSET1_PLL1_PS_SHIFT;
	ulX1FBD = (ulCLKSET & SYSCON_CLKSET1_PLL1_X1FBD1_MASK) >> SYSCON_CLKSET1_PLL1_X1FBD1_SHIFT;
	ulX2FBD = (ulCLKSET & SYSCON_CLKSET1_PLL1_X2FBD2_MASK) >> SYSCON_CLKSET1_PLL1_X2FBD2_SHIFT;
	ulX2IPD = (ulCLKSET & SYSCON_CLKSET1_PLL1_X2IPD_MASK) >> SYSCON_CLKSET1_PLL1_X2IPD_SHIFT;
                                                                                                                             
	/*
	 * This may be off by a very small fraction of a percent.
	 */
	ulPLL_Freq = (((0x00e10000 * (ulX1FBD+1)) / (ulX2IPD+1)) * (ulX2FBD+1)) >> ulPS;
                                                                                                                             
	return ulPLL_Freq;
}

/*
 * SetSampleRate
 * 		disables i2s channels and sets up i2s divisors
 * 		in syscon for the requested sample rate.
 *
 * lFrequency - Sample Rate in Hz
 */
static void ep93xx_set_samplerate(long lFrequency)
{
	ulong ulRequestedMClkFreq, ulPLL1_CLOCK, ulPLL2_CLOCK;
	ulong ulMClkFreq1, ulMClkFreq2, ulClkSet1, ulClkSet2;
	ulong ulI2SDiv, ulI2SDiv1, ulI2SDiv2, ulI2SDIV, actual_samplerate;
    
	/*
	 * Clock ratios: MCLK to SCLK and SCLK to LRCK
	 */
	ulong ulM2SClock  = 4;
	ulong ulS2LRClock = 64;

	DPRINTK( "ep93xx_set_samplerate = %d Hz.\n", (int)lFrequency );

	/*
	 * Read CLKSET1 and CLKSET2 in the System Controller and calculate
	 * the PLL frequencies from that.
	 */
	ulClkSet1 =	inl(SYSCON_CLKSET1);
	ulClkSet2 =	inl(SYSCON_CLKSET2);
	ulPLL1_CLOCK = ep93xx_get_PLL_frequency( ulClkSet1 );
	ulPLL2_CLOCK = ep93xx_get_PLL_frequency( ulClkSet2 );

	DPRINTK( "ClkSet1=0x%08x Clkset2=0x%08x  PLL1=%d Hz  PLL2=%d Hz\n", 
		(int)ulClkSet1, (int)ulClkSet2, (int)ulPLL1_CLOCK, (int)ulPLL2_CLOCK );

	ulRequestedMClkFreq = ( lFrequency * ulM2SClock * ulS2LRClock);

	ep93xx_calc_closest_freq
	(
		ulPLL1_CLOCK, 
		ulRequestedMClkFreq,
		&ulMClkFreq1,
		&ulI2SDiv1
	);
	ep93xx_calc_closest_freq
	(
		ulPLL2_CLOCK, 
		ulRequestedMClkFreq,
		&ulMClkFreq2,
		&ulI2SDiv2
	);

	/*
	 * See which is closer, MClk rate 1 or MClk rate 2.
	 */
	if(abs(ulMClkFreq1 - ulRequestedMClkFreq) < abs(ulMClkFreq2 -ulRequestedMClkFreq))
	{
		ulI2SDiv = ulI2SDiv1;
		actual_samplerate = ulMClkFreq1/ (ulM2SClock * ulS2LRClock);
		DPRINTK( "ep93xx_set_samplerate - using PLL1\n" );
	}
	else
	{
		ulI2SDiv = ulI2SDiv2 | SYSCON_I2SDIV_PSEL;
		actual_samplerate = ulMClkFreq1 / (ulM2SClock * ulS2LRClock);
		DPRINTK( "ep93xx_set_samplerate - using PLL2\n" );
	}

	/*
	 * Calculate the new I2SDIV register value and write it out.
	 */
	ulI2SDIV = ulI2SDiv | SYSCON_I2SDIV_SENA |  SYSCON_I2SDIV_ORIDE | 
				SYSCON_I2SDIV_SPOL| SYSCON_I2SDIV_LRDIV_64 | 
				SYSCON_I2SDIV_SDIV | SYSCON_I2SDIV_MENA | SYSCON_I2SDIV_ESEL;

	DPRINTK("I2SDIV set to 0x%08x\n", (unsigned int)ulI2SDIV );
	
	SysconSetLocked(SYSCON_I2SDIV, ulI2SDIV);
	audio_clocks_enabled_in_syscon = 1;

}

/*
 * BringUpI2S
 *
 * This routine sets up the I2S Controller.
 */
static void ep93xx_init_i2s_controller( void )
{
	unsigned int uiDEVCFG;
	
	DPRINTK("ep93xx_init_i2s_controller - enter\n");
	
	/*
	 * Configure the multiplexed Ac'97 pins to be I2S.  Also configure 
	 * EGPIO[4,5,6,13] to be SDIN's and SDOUT's for the second and third
	 * I2S stereo channels if the codec is a 6 channel codec.
	 */
	uiDEVCFG = inl(SYSCON_DEVCFG);
	
#ifdef CONFIG_CODEC_CS4271
	uiDEVCFG |= SYSCON_DEVCFG_I2SonAC97;
#else /* CONFIG_CODEC_CS4228A */
	uiDEVCFG |= SYSCON_DEVCFG_I2SonAC97 | SYSCON_DEVCFG_A1onG | SYSCON_DEVCFG_A2onG;
#endif
	    
	SysconSetLocked(SYSCON_DEVCFG, uiDEVCFG);

	/* Configure I2S Tx channel */

	/*
	 * Justify Left, MSB first	
	*/
	outl( 0, I2STXLinCtrlData );

	/*
	 * WL = 24 bits 
	 */
	outl( 1, I2STXWrdLen );

	/*
	 * Set the I2S control block to master mode.
	 * Tx bit clk rate determined by word legnth 
	 * Do gate off last 8 clks (24 bit samples in a 32 bit field)
	 * LRclk = I2S timing; LRck = 0 for left
	 */
	outl( (i2s_txcc_mstr | i2s_txcc_trel), I2STxClkCfg );
	
	/* Configure I2S rx channel */
    
	/*
	 * First, clear all config bits.
	 */
	outl( 0, I2SRXLinCtrlData );

	/*
	 * WL = 24 bits 
	 */
	outl( 1, I2SRXWrdLen );

	/* 
	 * Set the I2S control block to master mode.
	 * Rx bit clk rate determined by word legnth 
	 * Do gate off last 8 clks 
	 * setting i2s_rxcc_rrel gives us I2S timing
	 * clearing i2s_rlrs gives us LRck = 0 for left, 1 for right
	 * setting i2s_rxcc_nbcg specifies to not gate off extra 8 clocks 
	 */
	outl( (i2s_rxcc_bcr_64x | i2s_rxcc_nbcg |i2s_rxcc_mstr | i2s_rxcc_rrel), I2SRxClkCfg );

	/* 
	 * Do an input to push the outl's past the wrapper 
	 */
	uiDEVCFG = inl(SYSCON_DEVCFG);
    
	DPRINTK("ep93xx_init_i2s_controller - EXIT\n");
}

/*
 * ep93xx_init_i2s_codec
 *
 * Note that codec must be getting i2s clocks for any codec
 * register writes to work.
 */
static void ep93xx_init_i2s_codec( void )
{
#if defined(CONFIG_ARCH_EDB9301) || defined(CONFIG_ARCH_EDB9302)
	unsigned int uiPADR, uiPADDR;
#endif

#if defined(CONFIG_ARCH_EDB9315A)
	unsigned int uiPBDR, uiPBDDR;
#endif

	DPRINTK("ep93xx_init_i2s_codec - enter\n");
    
	if( !audio_clocks_enabled_in_syscon )
	{
		DPRINTK("ep93xx_init_i2s_codec - EXIT - clocks not enabled\n");
		return;
	}

#ifdef CONFIG_CODEC_CS4271

	/*
	 * Some EDB9301 boards use EGPIO1 (port 1, bit 1) for the I2S reset.
	 * EGPIO1 has a pulldown so if it isn't configured as an output, it is low.
	 */
#if defined(CONFIG_ARCH_EDB9301) || defined(CONFIG_ARCH_EDB9302)
	uiPADR  = inl(GPIO_PADR);
	uiPADDR = inl(GPIO_PADDR);
    
	/* Clear bit 1 of the data register */
	outl( (uiPADR & 0xfd), GPIO_PADR );
	uiPADR  = inl(GPIO_PADR);

	/* Set bit 1 of the DDR to set it to output
	 * Now we are driving the reset pin low.
	 */
	outl( (uiPADDR | 0x02), GPIO_PADDR );
	uiPADDR = inl(GPIO_PADDR);

	udelay( 2 );  /* plenty of time */

	/* Set bit 1 of the data reg.  Now we drive the reset pin high. */
	outl( (uiPADR | 0x02),  GPIO_PADR );
	uiPADR  = inl(GPIO_PADR);
#endif

#if defined(CONFIG_ARCH_EDB9315A)
    
	uiPBDR  = inl(GPIO_PBDR);
	uiPBDDR = inl(GPIO_PBDDR);
    
	/* Clear bit 1 of the data register */
	outl( (uiPBDR & 0xbf), GPIO_PBDR );
	uiPBDR  = inl(GPIO_PBDR);

	/* Set bit 1 of the DDR to set it to output
	 * Now we are driving the reset pin low.
	 */
	outl( (uiPBDDR | 0x40), GPIO_PBDDR );
	uiPBDDR = inl(GPIO_PBDDR);

	udelay( 2 );  /* plenty of time */

	/* Set bit 1 of the data reg.  Now we drive the reset pin high. */
	outl( (uiPBDR | 0x40),  GPIO_PBDR );
	uiPBDR  = inl(GPIO_PBDR);
#endif

	/*
	 * Write to the control port, setting the enable control port bit
	 * so that we can write to the control port.  OK?
	 */
	SSPDriver->Write( SSP_Handle, 7, 0x02 );

	/* Select slave, 24Bit I2S serial mode */
	SSPDriver->Write( SSP_Handle, 1, 0x01 );
	SSPDriver->Write( SSP_Handle, 6, 0x10 );

	/* Set AMUTE (auto-mute) bit. */
	SSPDriver->Write( SSP_Handle, 2, 0x00 );

#else // CONFIG_CODEC_CS4228A

	/* Nothing needs to happen here */

#endif
    
	DPRINTK("ep93xx_init_i2s_codec - EXIT\n");
}

/*
 * ep93xx_automute_i2s_codec
 *
 * Note that codec must be getting i2s clocks for any codec
 * register writes to work.
 */
static void ep93xx_automute_i2s_codec( void )
{
	DPRINTK("ep93xx_automute_i2s_codec - enter\n");

	if( !audio_clocks_enabled_in_syscon )
	{
		DPRINTK("ep93xx_automute_i2s_codec - EXIT - clocks not enabled\n");
		return;
	}

#ifdef CONFIG_CODEC_CS4271

	/*
	 * The automute bit is set by default for the CS4271.
	 * Clear the driver's mute flag and use the set_volume routine
	 * to write the current volume out with the mute bit cleared.
	 */
	giDAC_Mute = 0;
	ep93xx_set_volume( giOSS_Volume );

#else // CONFIG_ARCH_CS4228A

	SSPDriver->Write( SSP_Handle, 4, 0 );

	/* Unmute the DACs */
	SSPDriver->Write( SSP_Handle, 5, 0xF0 );

	/* Unmute the MUTEC pin, turn on automute. */
	SSPDriver->Write( SSP_Handle, 5, 0x40 );
    
	giDAC_Mute = 0;
    
#endif

	DPRINTK("ep93xx_automute_i2s_codec - EXIT\n");
}

static int ep93xx_set_volume( int iOSS_Volume )
{
	int iOSS_Volume_L, iOSS_Volume_R;

#ifdef CONFIG_CODEC_CS4271
	int iMute = 0;
#endif
    
	if( !audio_clocks_enabled_in_syscon )
	{
		return -EINVAL;
	}
    
	giOSS_Volume = iOSS_Volume;

	iOSS_Volume_L =  iOSS_Volume & 0x00ff;
	iOSS_Volume_R = (iOSS_Volume & 0xff00) >> 8;

	if (iOSS_Volume_L > 100)
		iOSS_Volume_L = 100;
	if (iOSS_Volume_L < 0)
		iOSS_Volume_L = 0;
	if (iOSS_Volume_R > 100)
		iOSS_Volume_R = 100;
	if (iOSS_Volume_R < 0)
		iOSS_Volume_R = 0;

#ifdef CONFIG_CODEC_CS4271
	/*
	 * CS4271 volume is 0 to -127 dB in 1 dB increments.
	 * So scale the volume so a range of 0 to 100 becomes 127 to 0.
	 */
	giDAC_Volume_L = 127 - ((iOSS_Volume_L * 127) / 100);
	giDAC_Volume_R = 127 - ((iOSS_Volume_R * 127) / 100);

	if( giDAC_Mute )
	{
		iMute = 0x80;
	}

	/*
	 * Write the volume for DAC1 and DAC2 (reg 4 and 5)
	 */
	SSPDriver->Write( SSP_Handle, 4, (giDAC_Volume_L | iMute) );
	SSPDriver->Write( SSP_Handle, 5, (giDAC_Volume_R | iMute) );

#else // CONFIG_CODEC_CS4228A

	/*
	 * CS4228A DAC attenuation is 0 to 90.5 dB in 0.5 dB steps so scale
	 * volume so that a value of 0 to 100 is scaled to 181 to 0
	 */
	giDAC_Volume_L = 181 - ((iOSS_Volume_L * 181) / 100);
	giDAC_Volume_R = 181 - ((iOSS_Volume_R * 181) / 100);

	/*
	 * Write the volume for DAC1 and DAC2 (reg 7 and 8)
	 */
	SSPDriver->Write( SSP_Handle, 7, giDAC_Volume_L );
	SSPDriver->Write( SSP_Handle, 8, giDAC_Volume_R );
	
#endif

	return 0;
}

/*
 * ep93xx_audio_init
 * Note that the codec powers up with its DAC's muted and
 * the serial format set to 24 bit I2S mode.
 */
static void ep93xx_audio_init(void *dummy)
{
	/*
	 * Mute the DACs. Disable all audio channels.  
	 * Must do this to change sample rate.
	 */
	ep93xx_audio_disable( 0 );

	/*
	 * Set up the i2s clocks in syscon.  Enable them.
	 */ 
	ep93xx_set_samplerate( AUDIO_SAMPLE_RATE_DEFAULT );

	/*
	 * Set i2s' controller serial format, and enable 
	 * the i2s controller.
	 */
	ep93xx_init_i2s_controller();
	
	/*
	 * Initialize codec serial format, etc.
	 */
	ep93xx_init_i2s_codec();

	/*
	 * Clear the fifo and enable the tx0 channel.
	 */
	ep93xx_audio_enable();

	/*
	 * Set the volume for the first time.
	 */
	ep93xx_set_volume( AUDIO_DEFAULT_VOLUME );
	
	/*
	 * Unmute the DAC and set the mute pin MUTEC to automute.
	 */
	ep93xx_automute_i2s_codec();
	
	audio_has_been_initialized = 1;
	
	DPRINTK("ep93xx_audio_init - EXIT\n");
}

static void print_audio_format( long format )
{
	switch( format )
	{
		case SNDRV_PCM_FORMAT_U8:		   
			DPRINTK( "AFMT_U8\n" );		   
			break;

		case SNDRV_PCM_FORMAT_S8:
			DPRINTK( "AFMT_S8\n" );		   
			break;

		case SNDRV_PCM_FORMAT_S16_LE:
			DPRINTK( "AFMT_S16_LE\n" );		   
			break;

		case SNDRV_PCM_FORMAT_S16_BE:
			DPRINTK( "AFMT_S16_BE\n" );		   
			break;

		case SNDRV_PCM_FORMAT_U16_LE:
			DPRINTK( "AFMT_U16_LE\n" );		   
			break;

		case SNDRV_PCM_FORMAT_U16_BE:
		default:
			DPRINTK( "AFMT_U16_BE\n" );		   
			break;
	}
}

/*
 * We convert to 24 bit samples that occupy 32 bits each.
 * Formats we support:
 *
 * AFMT_U8		   
 * AFMT_S16_LE	   	Little endian signed 16
 * AFMT_S8		   
 * AFMT_U16_LE	   	Little endian U16
 * AFMT_S32_BLOCKED	32 bit little endian format, taken from the rme96xx driver.
 *
 */
static void audio_set_format( audio_stream_t * s, long val )
{
	DPRINTK( "audio_set_format enter.  Format requested (%d) %d ", 
				(int)val,SNDRV_PCM_FORMAT_S16_LE);
	print_audio_format( val );
	
	switch( val )
	{
		case SNDRV_PCM_FORMAT_U8:		   
			s->audio_format = SNDRV_PCM_FORMAT_U8;
			s->audio_stream_bitwidth = 8;
			break;

		case SNDRV_PCM_FORMAT_S8:
			s->audio_format = SNDRV_PCM_FORMAT_S8;
			s->audio_stream_bitwidth = 8;
			break;

		case SNDRV_PCM_FORMAT_S32_BE:
		case SNDRV_PCM_FMTBIT_S32_LE:
		
			s->audio_format = SNDRV_PCM_FORMAT_S32_LE;
			s->audio_stream_bitwidth = 32;
			break;

		case SNDRV_PCM_FORMAT_S16_LE:
		case SNDRV_PCM_FORMAT_S16_BE:
			s->audio_format = SNDRV_PCM_FORMAT_S16_LE /*AFMT_S16_LE*/;
			s->audio_stream_bitwidth = 16;
			break;

		case SNDRV_PCM_FORMAT_U16_LE:
		case SNDRV_PCM_FORMAT_U16_BE:
		default:
			s->audio_format = SNDRV_PCM_FORMAT_U16_LE/*AFMT_U16_LE*/;
			s->audio_stream_bitwidth = 16;
			break;
	}

	DPRINTK( "audio_set_format EXIT format set to be (%d) ", (int)s->audio_format );
	print_audio_format( (long)s->audio_format );
}

static __inline__ unsigned long copy_to_user_U16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	short * user_ptr = (short *)to;	/* 16 bit user buffer */
	int * dma_ptr = (int *)from;	/* 32 bit dma buffer  */

	int shift = 8;
	
	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( ((short)( (*dma_ptr++) >> shift )) ^ 0x8000, user_ptr++ );
			__put_user( ((short)( (*dma_ptr++) >> shift )) ^ 0x8000, user_ptr++ );
			to_count -= 4;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( ((short)( (*dma_ptr++) >> shift )) ^ 0x8000, user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count -= 2;
		}
	}
	
	return total_to_count;
}

static __inline__ unsigned long copy_to_user_S16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	short * user_ptr = (short *)to;	/* 16 bit user buffer */
	int * dma_ptr = (int *)from;	/* 32 bit dma buffer  */

	int shift = 8;
	
	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( (short)( (*dma_ptr++) >> shift ), user_ptr++ );
			__put_user( (short)( (*dma_ptr++) >> shift ), user_ptr++ );
			to_count -= 4;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( (short)((*dma_ptr++) >> shift), user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count -= 2;
		}
	}
	
	return total_to_count;
}

static __inline__ unsigned long copy_to_user_S8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	char * user_ptr = (char *)to;  /*  8 bit user buffer */
	int * dma_ptr = (int *)from;    /* 32 bit dma buffer  */

	int shift = 16;

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( (char)( (*dma_ptr++) >> shift ), user_ptr++ );
			__put_user( (char)( (*dma_ptr++) >> shift ), user_ptr++ );
			to_count -= 2;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( (char)((*dma_ptr++) >> shift), user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count--;
		}
	}
	
	return total_to_count;
}

static __inline__ unsigned long copy_to_user_U8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	char * user_ptr = (char *)to;  /*  8 bit user buffer */
	int * dma_ptr = (int *)from;    /* 32 bit dma buffer  */

	int shift = 16;

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( ((char)( (*dma_ptr++) >> shift )) ^ 0x80, user_ptr++ );
			__put_user( ((char)( (*dma_ptr++) >> shift )) ^ 0x80, user_ptr++ );
			to_count -= 2;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( ((char)( (*dma_ptr++) >> shift )) ^ 0x80, user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count--;
		}
	}
	
	return total_to_count;
}

/*
 * Returns negative for error
 * Returns # of bytes transferred out of the from buffer
 * for success.
 */
static __inline__ int copy_to_user_with_conversion
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount
)
{
	int ret=0;
	
	if( toCount == 0 )
	{
		DPRINTK("copy_to_user_with_conversion - nothing to copy!\n");
	}

	switch( state->input_stream->audio_format )
	{
		case SNDRV_PCM_FORMAT_U8:
			ret = copy_to_user_U8( state, to, from, toCount );
			break;

		case SNDRV_PCM_FORMAT_S16_LE:
			ret = copy_to_user_S16_LE( state, to, from, toCount );
			break;
		
		case SNDRV_PCM_FORMAT_S8:
			ret = copy_to_user_S8( state, to, from, toCount );
			break;
			
		case SNDRV_PCM_FORMAT_U16_LE:
			ret = copy_to_user_U16_LE( state, to, from, toCount );
			break;
				
		default:
			__copy_to_user( (char *)to, from, toCount);
			ret = toCount;
			break;
		}

	return ret;
}

static __inline__ int copy_from_user_U16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	int * dma_buffer = (int *)to;
	int val;
	unsigned short * user_buffer = (unsigned short *)from;
	unsigned short data;
	
	int toCount0 = toCount;
	int mono = (state->output_stream->audio_num_channels == 1);
	int shift = 8;

	if (!mono) 
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			*dma_buffer++ = ((unsigned int)data ^ 0x8000) << shift;
			__get_user(data, user_buffer++);
			*dma_buffer++ = ((unsigned int)data ^ 0x8000) << shift;
			toCount -= 8;
		}
	}
	else
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			val = ((unsigned int)data ^ 0x8000) << shift;
			*dma_buffer++ = val;
			*dma_buffer++ = val;
			toCount -= 8;
		}
	}
	
	/*
	 * Each mono or stereo 16 bit sample going in results
	 * in a two 32 bit (left & right) samples going out.  So
	 * mono   = (2 bytes in)/(8 bytes out)
	 * stereo = (4 bytes in)/(8 bytes out)
	 */
	if( mono )
		return toCount0 / 4;
	
	return toCount0 / 2;
}

static __inline__ int copy_from_user_S16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	int * dma_buffer = (int *)to;
	int val;
	unsigned short * user_buffer = (unsigned short *)from;
	unsigned short data;
	
	int toCount0 = toCount;
	int mono = (state->output_stream->audio_num_channels == 1);
	int shift = 8;

	if (!mono)
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			*dma_buffer++ = (unsigned int)data << shift;
			__get_user(data, user_buffer++);
			*dma_buffer++ = (unsigned int)data << shift;
			toCount -= 8;
		}
	}
	else
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			val = (unsigned int)data << shift;
			*dma_buffer++ = val;
			*dma_buffer++ = val;
			toCount -= 8;
		}
	} 

	/*
	 * Each mono or stereo 16 bit sample going in results
	 * in a two 32 bit samples (left & right) going out.  So
	 * mono   = (2 bytes in)/(8 bytes out)
	 * stereo = (4 bytes in)/(8 bytes out)
	 */
	if( mono )
		return toCount0 / 4;
	
	return toCount0 / 2;
}

static __inline__ int copy_from_user_S8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	int * dma_buffer = (int *)to;
	int val;
	unsigned char * user_buffer = (unsigned char *)from;
	unsigned char data;
	
	int toCount0 = toCount;
	int mono = (state->output_stream->audio_num_channels == 1);
	int shift = 16;

	if (!mono) 
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			*dma_buffer++ = (unsigned int)data << shift;
			__get_user(data, user_buffer++);
			*dma_buffer++ = (unsigned int)data << shift;
			toCount -= 8;
		}
	}
	else
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			val = (unsigned int)data << shift;
			*dma_buffer++ = val;
			*dma_buffer++ = val;
			toCount -= 8;
		}
	}
	

	/*
	 * Each mono or stereo 8 bit sample going in results
	 * in a two 32 bit samples (left & right) going out.  So
	 * mono   = (1 byte  in)/(8 bytes out)
	 * stereo = (2 bytes in)/(8 bytes out)
	 */
	if( mono )
		return toCount0 / 8;
	
	return toCount0 / 4;
}

static __inline__ int copy_from_user_U8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	unsigned int * dma_buffer = (unsigned int *)to;
	unsigned int val;
	unsigned char * user_buffer = (unsigned char *)from;
	unsigned char data;
	
	int toCount0 = toCount;
	int mono = (state->output_stream->audio_num_channels == 1);
	int shift = 16;

	if (!mono)
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			*dma_buffer++ = ((unsigned int)data ^ 0x80) << shift;
			__get_user(data, user_buffer++);
			*dma_buffer++ = ((unsigned int)data ^ 0x80) << shift;
			toCount -= 8;
		}
	}
	else
	{
		while (toCount > 0) 
		{
			__get_user(data, user_buffer++);
			val = ((unsigned int)data ^ 0x80) << shift;
			*dma_buffer++ = val;
			*dma_buffer++ = val;
			toCount -= 8;
		}
	}
	

	/*
	 * Each mono or stereo 8 bit sample going in results
	 * in a two 32 bit samples (left & right) going out.  So
	 * mono   = (1 byte  in)/(8 bytes out)
	 * stereo = (2 bytes in)/(8 bytes out)
	 */
	if( mono )
		return toCount0 / 8;
	
	return toCount0 / 4;
}

/*
 * Returns negative for error
 * Returns # of bytes transferred out of the from buffer
 * for success.
 */
static __inline__ int copy_from_user_with_conversion
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	int ret=0;
	//DPRINTK("copy_from_user_with_conversion\n");	
	if( toCount == 0 )
	{
		DPRINTK("copy_from_user_with_conversion - nothing to copy!\n");
	}

	switch( state->output_stream->audio_format )
	{
		case SNDRV_PCM_FORMAT_U8:
			//DPRINTK("SNDRV_PCM_FORMAT_U8\n");
			ret = copy_from_user_U8( state, to, from, toCount );
			break;

		case SNDRV_PCM_FORMAT_S16_LE:
			//DPRINTK("SNDRV_PCM_FORMAT_S16_LE\n");
			ret = copy_from_user_S16_LE( state, to, from, toCount );
			break;
		
		case SNDRV_PCM_FORMAT_S8:
			//DPRINTK("SNDRV_PCM_FORMAT_S8\n");
			ret = copy_from_user_S8( state, to, from, toCount );
			break;
				
		case SNDRV_PCM_FORMAT_U16_LE:
			//DPRINTK("SNDRV_PCM_FORMAT_U16_LE\n");
			ret = copy_from_user_U16_LE( state, to, from, toCount );
			break;
				
		default:
			__copy_from_user( (char *)to, from, toCount);
			ret = toCount;
			break;
	}
	
	return ret;
}

/*
 *  audio_dma2usr_ratio()
 *
 *  For audio playback, we convert samples of arbitrary format to be 32 bit 
 *  for our hardware. We're scaling a user buffer to a dma buffer.  So when
 *  report byte counts, we scale them acording to the ratio of DMA sample
 *  size to user buffer sample size.  When we report # of DMA fragments,
 *  we don't scale that.  So:
 *   - The fragment size the app sees  = (stream->fragsize/stream->dma2usr_ratio)
 *   - The # of fragments the app sees = stream->nbfrags
 *
 *	User sample type can be stereo/mono/8/16/32 bit.  
 *  DMA sample type will be either CM (compact mode) where two 16 bit 
 *  samples together in a 32 bit word become a stereo sample or non-CM 
 *  where each channel gets a 32 bit word.
 *
 *  Any time usr sample type changes, we need to call this function.
 *
 *  Also adjust the size and number of dma fragments if sample size changed.
 *
 *  Input format       Input sample     Output sample size    ratio (out:in)
 *  bits   channels    size (bytes)       CM   non-CM          CM   non-CM
 *   8       mono          1               4      8            4:1   8:1
 *   8      stereo         2		   4      8            2:1   4:1
 *   16      mono          2		   4      8            2:1   4:1
 *   16     stereo         4		   4      8            1:1   2:1
 *
 *   24      mono          3		   4      8             X    8:3 not a real case
 *   24     stereo         6		   4      8             X    8:6 not a real case
 *   32      mono          4		   4      8             X    2:1
 *   32     stereo         8		   4      8             X    1:1
 *
 */
static void audio_dma2usr_ratio( audio_stream_t * stream )
{
	unsigned int dma_sample_size, user_sample_size;
	
	dma_sample_size = 8;	/* each stereo sample is 2 * 32 bits */
	
	if( stream->audio_num_channels != 1 )
	{
		// If stereo 16 bit, user sample is 4 bytes.
		// If stereo  8 bit, user sample is 2 bytes.
		user_sample_size = stream->audio_stream_bitwidth / 4;
	}
	else
	{
		// If mono 16 bit, user sample is 2 bytes.
		// If mono  8 bit, user sample is 1 bytes.
		user_sample_size = stream->audio_stream_bitwidth / 8;
	}

	/* dma2usr_ratio = (4 or 8) / (4, 2, or 1) = 8, 4, 2, or 1 */
	stream->dma2usr_ratio = dma_sample_size / user_sample_size;
	DPRINTK("stream->dma2usr_ratio = %d\n",stream->dma2usr_ratio);
}

/*---------------------------------------------------------------------------------------------*/

/*
 * audio_dma_start
 *
 * Our Ac97 has a specific start order that it likes.  Enable the 
 * Ac97 channel AFTER enabling DMA.  Our I2S is not so picky.
 */
static void audio_dma_start( audio_state_t * state, audio_stream_t * stream )
{
	int err;

	DPRINTK("audio_dma_start\n");

	err = ep93xx_dma_start( stream->dmahandles[0],
				stream->NumDmaChannels, 
				stream->dmahandles );
	if (state->hw_enable){
		state->hw_enable( state->output_stream == stream );
	}
}

/*
 * audio_dma_pause
 */
static void audio_dma_pause( audio_state_t * state, audio_stream_t * stream )
{
	DPRINTK("audio_dma_pause\n");
	if (state->hw_disable){
		state->hw_disable( state->output_stream == stream );
	}
	
	ep93xx_dma_pause( stream->dmahandles[0],
			  stream->NumDmaChannels, 
			  stream->dmahandles );
}

/*
 * audio_deallocate_buffers
 * 
 * This function frees all buffers
 */
static void audio_deallocate_buffers( snd_pcm_substream_t *substream, audio_stream_t *s )
{
	int frag, i;
	audio_state_t  *state = (audio_state_t *)substream->private_data;
	
	DPRINTK("EP93xx - audio_deallocate_buffers\n");

	/* ensure DMA won't run anymore */
	audio_dma_pause( state, s );
	s->active = 0;
	s->stopped = 0;

	for( i=0 ; i<s->NumDmaChannels ; i++ )
	{
		ep93xx_dma_flush( s->dmahandles[0] );
	}
	
	if (s->buffers) 
	{
		for (frag = 0; frag < s->nbfrags; frag++) 
		{
			if (!s->buffers[frag].master)
			{
				continue;
			}

			s->buffers[frag].start = NULL;
			s->buffers[frag].master = 0;
			s->buffers[frag].dma_addr = (dma_addr_t)NULL;			
		}

		/* Free the space allocated to the array of buf structs. */
		kfree(s->buffers);
		s->buffers = NULL;
	}

	s->buf_idx = 0;
	s->buf = NULL;

	DPRINTK("EP93xx - audio_deallocate_buffers - EXIT\n");
}


/*
 * audio_allocate_buffers
 *
 * This function allocates the buffer structure array and buffer data space
 * according to the current number of fragments and fragment size.
 * Note that the output_stream and input_stream structs are allocated
 * in ep93xx-ac97.c or ep93xx-i2s.c.
 */
static int audio_allocate_buffers( snd_pcm_substream_t *substream, audio_stream_t * s)
{
	int frag;
	int dmasize = 0;
	char *dmabuf = NULL;
	dma_addr_t dmaphys = 0;
	
	struct snd_dma_buffer *pcmbuf = &substream->dma_buffer;
	
	if (s->buffers){
		DPRINTK("%s BUSY\n",__FUNCTION__);
		return -EBUSY;
	}

//	DPRINTK("EP93xx audio_allocate_buffers ENTER\n" );
	
	/* Allocate space for the array of audio_buf_t structs. */
	s->buffers = (audio_buf_t *)
		kmalloc(sizeof(audio_buf_t) * s->nbfrags, GFP_KERNEL);

	if (!s->buffers)
	{
		DPRINTK(AUDIO_NAME ": unable to allocate audio memory\n ");
		audio_deallocate_buffers(substream,s);
		return -ENOMEM;
	}

	memset( s->buffers, 0, sizeof(audio_buf_t) * s->nbfrags );

	/*
	 * Let's allocate non-cached memory for DMA buffers.
	 * We try to allocate all memory at once.
	 * If this fails (a common reason is memory fragmentation),
	 * then we allocate more smaller buffers.
	 */
	
	for (frag = 0; frag < s->nbfrags; frag++) 
	{
		audio_buf_t *buf = &s->buffers[frag];

		if (!dmasize) 
		{
			dmasize	= s->externedfragsize * s->nbfrags;
			DPRINTK("the s->externedfragsize = %x\n",s->externedfragsize);
			DPRINTK("the s->nbfrags = %x\n",s->nbfrags);
			DPRINTK("the dmasize = %x\n",dmasize);

			do {
				dmabuf	= pcmbuf->area;
				dmaphys	= pcmbuf->addr;
			} while (!dmabuf && dmasize);
		
			/*
			 * If we do fail to allocate enough for all the frags, 
			 * deallocate whatever we did get and quit.
			 */
			if (!dmabuf)
			{
				DPRINTK(AUDIO_NAME ": unable to allocate audio memory\n ");
				audio_deallocate_buffers(substream, s);
				return -ENOMEM;
			}

			DPRINTK("EP93xx allocated %x bytes:  dmabuf=0x%08x  dmaphys=0x%08x\n",
					dmasize, (int)dmabuf, (int)dmaphys );
			
			/*
			 * Success!	 Save the size of this chunk to use when we deallocate it.
			 */
			buf->master = dmasize;
			memzero(dmabuf, dmasize);
		}

		/*
		 * Save the start address of the buf fragment.
		 * We know the size of the fragment is fragsize.
		 */
		buf->start = dmabuf;
		buf->dma_addr = dmaphys;
		DPRINTK("buf->start=%x,buf->dma_addr=%x\n",buf->start,buf->dma_addr);
		buf->stream = s;
		buf->size = 0;
		buf->reportedsize = s->fragsize;
		buf->sent = 0;

		sema_init(&buf->sem, 1);

		/*
		 * Now if we only allocated the minimal one frag of space, the
		 * dmasize will be ==0 after this subtraction so it will allocate more
		 * for the next frag.  Otherwise, the next time(s) thru this for loop
		 * will dole out frag sized pieces of this big master chunk.
		 */
		//dmabuf += s->fragsize;
		//dmaphys += s->fragsize;
		//dmasize -= s->fragsize;

		dmabuf += s->externedfragsize;
		dmaphys += s->externedfragsize;
		dmasize -= s->externedfragsize;
	}

	/*
	 * Initialize the stream.
	 */
	s->buf_idx = 0;			/* Init the current buffer index. */
	s->buf = &s->buffers[0];	/* Point buf to the current buffer struct. */
	s->bytecount = 0;
	s->externedbytecount = 0;
	s->getptrCount = 0;
	s->fragcount = 0;

//	DPRINTK("EP93xx audio_allocate_buffers -- exit SUCCESS\n" );
	return 0;

}

/*
 * DMA callback functions
 */
static void audio_dma_tx_callback
( 
	ep93xx_dma_int_t DMAInt,
	ep93xx_dma_dev_t device, 
	unsigned int user_data 
)
{
	unsigned int buf_id;
	int handle;
	
	snd_pcm_substream_t *substream = (snd_pcm_substream_t *)user_data;
	audio_state_t *state = (audio_state_t *)(substream->private_data);
	audio_stream_t *s = state->output_stream;

	DPRINTK( "audio_dma_tx_callback  - enter\n"); 
	
	switch( device )
	{
    		case DMATx_I2S3:
			handle = s->dmahandles[2];
			break;

    		case DMATx_I2S2:
			handle = s->dmahandles[1];
			break;

		case DMATx_I2S1:
		default:
			handle = s->dmahandles[0];
			break;
	}
	
	/*
	 * If we are mapped, get one buffer back and recycle it.
	 */
	if( ep93xx_dma_remove_buffer( handle, &buf_id ) >= 0 )
	{
	   	audio_buf_t *buf = (audio_buf_t *)buf_id;
		
		/* Accounting */
		s->bytecount += s->fragsize;
		
		DPRINTK("s->bytecount=%x %x\n",s->bytecount,s->externedfragsize);

		/* Recycle buffer */
		if(s->stopped==0)
		{	
			snd_pcm_period_elapsed(substream);	
		 	ep93xx_dma_add_buffer( s->dmahandles[0], 	/* dma instance */
						(unsigned int)buf->dma_addr, /* source 	*/
						0, 			/* dest */
						s->externedfragsize,	/*s->fragsize,*/	/* size 	*/
						0,		/* is the last chunk? 	*/
						(unsigned int) buf ); /* buf id		*/	

		}
	}

}

static void audio_dma_rx_callback
(
	ep93xx_dma_int_t DMAInt,
	ep93xx_dma_dev_t device, 
	unsigned int user_data 
)
{
	unsigned int buf_id;
	int handle;

	snd_pcm_substream_t *substream	= (snd_pcm_substream_t *)user_data;
	audio_state_t *state = (audio_state_t *)(substream->private_data);
	audio_stream_t *s = state->input_stream;

	DPRINTK("audio_dma_rx_callback\n");
	switch( device )
	{
		case DMARx_I2S3:
			handle = s->dmahandles[2];
			break;

    		case DMARx_I2S2:
			handle = s->dmahandles[1];
			break;

		case DMARx_I2S1:
		default:
			handle = s->dmahandles[0];
			break;
	}

	/*
	 * Keep removing and recycling buffers as long as there are buffers
	 * to remove.
	 */
	while( !ep93xx_dma_remove_buffer( handle, &buf_id ) )
	{
	   	audio_buf_t *buf = (audio_buf_t *)buf_id;
   
		/* Accounting */
		s->bytecount += s->fragsize;

		DPRINTK("s->bytecount=%x %x\n",s->bytecount,s->externedfragsize);

		/* Recycle buffer */
		if(s->stopped==0)
		{	
			snd_pcm_period_elapsed(substream);	
		 	ep93xx_dma_add_buffer( s->dmahandles[0], 	/* dma instance */
						(unsigned int)buf->dma_addr, /* source 	*/
						0, 			/* dest */
						s->externedfragsize,	/*s->fragsize,*/	/* size 	*/
						0,		/* is the last chunk? 	*/
						(unsigned int) buf ); /* buf id		*/	

		}

 	} 

}

static int audio_release(snd_pcm_substream_t *substream)
{
	int i;
	audio_state_t *state = (audio_state_t *)substream->private_data;

	DPRINTK("EP93xx - audio_release -- enter.  \n");
			

	down(&state->sem);

	/*read data from codec to controller */
	if(substream->stream == SNDRV_PCM_STREAM_CAPTURE) 
	{
		audio_deallocate_buffers(substream, state->input_stream);
		
		for( i=0 ; i < state->input_stream->NumDmaChannels ; i++ )
		{
			ep93xx_dma_free( state->input_stream->dmahandles[i] );
		}
		state->rd_ref = 0;
	}
	
	/*write data from controller to codec*/
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK ) 
	{
		/*
		 * Wait for all our buffers to play.
		 */
		//audio_sync(substream);
		
		/*
		 * audio_deallocate_buffers will pause and flush the dma
		 * then deallocate the buffers.
		 */
		audio_deallocate_buffers(substream, state->output_stream);
		for( i=0 ; i < state->output_stream->NumDmaChannels ; i++ )
		{
			ep93xx_dma_free( state->output_stream->dmahandles[i] );
		}
		state->wr_ref = 0;
	}

	up(&state->sem);

	DPRINTK("EP93xx - audio_release -- EXIT\n");

	return 0;
}


/*
 *ep93xx_i2s_pcm_hw_params
 *set the pcm dma buffer to the runtime pcm dma buffer
 */
static int ep93xx_i2s_pcm_hw_params(snd_pcm_substream_t *substream,
				snd_pcm_hw_params_t *params)
{
	DPRINTK("ep93xx_i2s_pcm_hw_params----ENTER\n");
	
	return snd_pcm_lib_malloc_pages(substream,params_buffer_bytes(params));	
}

/*
 *free the runtime_data buffer 
 */
static int ep93xx_i2s_pcm_hw_free(snd_pcm_substream_t *substream)
{
	DPRINTK("ep93xx_i2s_pcm_hw_free----ENTER\n");
	
	return snd_pcm_lib_free_pages(substream);
}

/*
 *ep93xx_i2s_pcm_prepare: need to finish these functions as lower
 *chip_set_sample_format
 *chip_set_sample_rate
 *chip_set_channels
 *chip_set_dma_setup
 */
static int ep93xx_i2s_pcm_prepare(snd_pcm_substream_t *substream)
{
	audio_state_t *state = (audio_state_t *) substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;
	audio_stream_t *stream = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
					state->output_stream : state->input_stream;

	DPRINTK("ep93xx_snd_pcm_prepare - ENTER\n");

	//DPRINTK("The runtime item : \n");
	//DPRINTK("runtime->dma_addr    = 0x%x\n", runtime->dma_addr);
	//DPRINTK("runtime->dma_area    = 0x%x\n", runtime->dma_area);
	//DPRINTK("runtime->dma_bytes   = 0x%x\n", runtime->dma_bytes);
	//DPRINTK("runtime->frame_bits  = 0x%x\n", runtime->frame_bits);
	//DPRINTK("runtime->buffer_size = 0x%x\n", runtime->buffer_size);
	//DPRINTK("runtime->period_size = 0x%x\n", runtime->period_size);
	//DPRINTK("runtime->periods     = 0x%x\n", runtime->periods);
	//DPRINTK("runtime->rate        = 0x%x\n", runtime->rate);
	//DPRINTK("runtime->format      = 0x%x\n", runtime->format);

	/* set requestd format when available */
	stream->audio_num_channels = runtime->channels;
	stream->fragsize = frames_to_bytes (runtime, runtime->period_size);
	stream->nbfrags = runtime->periods;

	/* set the audio format */
	audio_set_format( stream, runtime->format );

	/*
	 * Mute the DACs. Disable all audio channels.  
	 * Must do this to change sample rate.
	 */
	ep93xx_audio_disable( 0 );

	/*
	 * Set up the i2s clocks in syscon.  Enable them.
	 */ 
	ep93xx_set_samplerate( runtime->rate );

	/*
	 * Set i2s' controller serial format, and enable 
	 * the i2s controller.
	 */
	ep93xx_init_i2s_controller();
	
	/*
	 * Initialize codec serial format, etc.
	 */
	ep93xx_init_i2s_codec();

	/*
	 * Clear the fifo and enable the tx0 channel.
	 */
	ep93xx_audio_enable();

	/*
	 * Set the volume for the first time.
	 */
	ep93xx_set_volume( AUDIO_DEFAULT_VOLUME );
	
	/*
	 * Unmute the DAC and set the mute pin MUTEC to automute.
	 */
	ep93xx_automute_i2s_codec();
	
	audio_dma2usr_ratio( stream );

	/* set the stream->period_size to the value of dma space */
	stream->externedfragsize = stream->fragsize * stream->dma2usr_ratio;
	DPRINTK("stream->period_size = 0x%x\n",stream->externedfragsize);
	
	DPRINTK("ep93xx_snd_pcm_prepare - EXIT\n");
	return 0;	
}

/*
 *start/stop/pause dma translate
 */
static int ep93xx_i2s_pcm_trigger(snd_pcm_substream_t *substream, int cmd)
{	
	audio_state_t  *state = (audio_state_t *)substream->private_data;
	audio_stream_t *stream = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
					state->output_stream:state->input_stream;
	
	int ret = 0, i;
	unsigned long flags;

	switch (cmd)
	{
		case SNDRV_PCM_TRIGGER_START:
			if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
			{
				/*
		 	 	 * Allocate dma buffers if we haven't already done so.
		 		 */
				if( !stream->buffers && audio_allocate_buffers(substream, stream) )
				{
					DPRINTK("%s error %i\n", __FUNCTION__, __LINE__);
					return -ENOMEM;
				}
		
				/*
		 		 * If we haven't already started the DMA start it.
				 * But don't start it if we are waiting on a trigger.
				 */
				if( !stream->active && !stream->stopped )
				{
					stream->active = 1;
					audio_dma_start( state, stream );
				}

				/* Add the audio data to dma request */
				local_irq_save(flags);
				
				for (i = 0; i < stream->nbfrags; i++) 
				{
					audio_buf_t *buf = stream->buf;
	 				ep93xx_dma_add_buffer( stream->dmahandles[0],	/* dma instance */
							(unsigned int)buf->dma_addr, 	/* source */
							0, 				/* dest */
							stream->externedfragsize,	/* size	*/
							0,				/* is the last chunk? 	*/
							(unsigned int) buf ); 		/* buf id	*/	
					NEXT_BUF(stream, buf);
				}

				local_irq_restore(flags);
			}
			if(substream->stream == SNDRV_PCM_STREAM_CAPTURE)
			{
				/* Allocate dma buffers if we haven't already done so. */
				if( !stream->buffers && audio_allocate_buffers(substream, stream) )
				{
					DPRINTK("%s error %i\n", __FUNCTION__, __LINE__);
					return -ENOMEM;
				}
				
				/*
				 * If we haven't already started the DMA start it.
				 * But don't start it if we are waiting on a trigger.
				 */
				if( !stream->active && !stream->stopped )
				{
					stream->active = 1;
					audio_dma_start( state, stream );
				}

				local_irq_save(flags);
				for (i = 0; i < stream->nbfrags; i++) 
				{
					audio_buf_t *buf = stream->buf;
	 				ep93xx_dma_add_buffer( stream->dmahandles[0],	/* dma instance */
							(unsigned int)buf->dma_addr, 	/* source */
							0, 				/* dest */
							stream->externedfragsize,	/* size	*/
							0,				/* is the last chunk? 	*/
							(unsigned int) buf ); 		/* buf id	*/	
					NEXT_BUF(stream, buf);
				}
				local_irq_restore(flags);
			}
			break;

		case SNDRV_PCM_TRIGGER_STOP:
			audio_dma_pause( state, stream );
			stream->stopped = 1;
			stream->active 	= 0;
			break;

		case SNDRV_PCM_TRIGGER_SUSPEND:
			break;
		case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
			break;
		case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
			break;

		default:
			ret = -EINVAL;
			DPRINTK("SNDRV_PCM_TRIGGER didnot support this command %x",cmd);
	}

	return ret;
}

/*
 * Get the frame location
 */
static snd_pcm_uframes_t ep93xx_i2s_pcm_pointer(snd_pcm_substream_t *substream)
{
	audio_state_t *state = (audio_state_t *)(substream->private_data);
	snd_pcm_runtime_t *runtime = substream->runtime;
	audio_stream_t *stream 	= (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
					state->output_stream:state->input_stream;
	snd_pcm_uframes_t pointer = 0;

 	pointer = bytes_to_frames(runtime,stream->bytecount);

	if (pointer >= runtime->buffer_size)
	{
  		pointer = 0;
	   	stream->bytecount=0;
		stream->externedbytecount=0;
	}
	return pointer;
}

static int ep93xx_i2s_pcm_open(snd_pcm_substream_t *substream)
{
	audio_state_t *state = substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;
	
	audio_stream_t *os = state->output_stream;
	audio_stream_t *is = state->input_stream;
	int err = 0;

	DPRINTK("ep93xx_i2s_pcm_open----ENTER\n");
	
	/* set the hardware description*/
	runtime->hw = ep93xx_i2s_pcm_hardware;


	down(&state->sem);
	
	/*write :output direction*/
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK ) 
	{
		DPRINTK("ep93xx_audio_open -- FMODE_WRITE\n");
		
		strncpy( os->devicename, state->output_id, MAX_DEVICE_NAME );

		/* 
	 	 * Request DMA channels 
	 	 */
		err = ep93xx_dma_request( 	&os->dmahandles[0], 
						os->devicename,
                       				state->output_dma );
		if (err)
		{
			up(&state->sem);
			DPRINTK("ep93xx_audio_open -- EXIT ERROR dma request failed\n");
			return err;
		}
		/* 
	 	 * Config DMA channels 
	 	 */
		err = ep93xx_dma_config(os->dmahandles[0], 
					IGNORE_CHANNEL_ERROR, 
					0, 
					audio_dma_tx_callback,
					(unsigned int)substream );
		if (err) 
		{
			ep93xx_dma_free( os->dmahandles[0] );
			up(&state->sem);
			DPRINTK("ep93xx_audio_attach -- EXIT ERROR dma config failed\n");
			return err;
		}

		os->NumDmaChannels = 1;
		state->wr_ref = 1;
		audio_deallocate_buffers(substream,os);

		os->fragsize 	= AUDIO_FRAGSIZE_DEFAULT;
		os->nbfrags  	= AUDIO_NBFRAGS_DEFAULT;

		os->reportedfrags = os->dma2usr_ratio * AUDIO_NBFRAGS_DEFAULT ;

		os->mapped = 0;
		os->stopped = 0;
		init_waitqueue_head(&os->wq);
	}
	
	/*read :input direction*/
	if(substream->stream == SNDRV_PCM_STREAM_CAPTURE) 
	{
		strncpy( is->devicename, state->input_id, MAX_DEVICE_NAME );
		
		/* 
	 	 * Request DMA channels 
	 	 */
		err = ep93xx_dma_request( 	&is->dmahandles[0], 
						is->devicename,
                       				state->input_dma);
		if (err) 
		{
			up(&state->sem);
			DPRINTK("ep93xx_audio_attach -- EXIT ERROR dma request failed\n");
			return err;
		}
		
		/* 
	 	 * Config DMA channels 
	 	 */
		err = ep93xx_dma_config( 	is->dmahandles[0], 
						IGNORE_CHANNEL_ERROR, 
						0,
						audio_dma_rx_callback, 
						(unsigned int)substream );
		if (err) 
		{
			ep93xx_dma_free( is->dmahandles[0] );
			up(&state->sem);
			DPRINTK("ep93xx_audio_attach -- EXIT ERROR dma config failed\n");
			return err;
		}

		is->NumDmaChannels = 1;
		state->rd_ref = 1;
		audio_deallocate_buffers(substream,is);
		is->fragsize 	= AUDIO_FRAGSIZE_DEFAULT;
		is->nbfrags 	= AUDIO_NBFRAGS_DEFAULT;

		is->reportedfrags = is->dma2usr_ratio * AUDIO_NBFRAGS_DEFAULT ;

		is->mapped = 0;
		is->stopped = 	0;
		init_waitqueue_head(&is->wq);
	}

	up(&state->sem);

	DPRINTK("ep93xx_audio_open -- EXIT\n");
	return err;
}

/*
 *free the HW dma channel
 *free the HW dma buffer
 *free the Hw dma decrotion using function :kfree
 *shut down the ac97 codec and  Ac97 controller
 */
static int ep93xx_i2s_pcm_close(snd_pcm_substream_t *substream)
{
	int ret;

	DPRINTK("ep93xx_i2s_pcm_close----ENTER\n");

	ret = audio_release(substream);
	if(ret!=0)
		DPRINTK("%s error %x",__FUNCTION__,ret);

	DPRINTK("ep93xx_i2s_pcm_close----EXIT\n");
	return 0;
}

static int ep93xx_i2s_pcm_playback_copy(snd_pcm_substream_t * substream,int channel, 
				snd_pcm_uframes_t pos,void __user *src, snd_pcm_uframes_t count)
{

	audio_state_t *state = (audio_state_t *)substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;
	audio_stream_t *stream = state->output_stream;
	int ret = 0;

	int tocount = frames_to_bytes(runtime,count);

	size_t dma_csize = runtime->dma_bytes / runtime->channels;
	char *hwbuf;

	if( channel == -1 )
	{
		hwbuf = runtime->dma_area  + stream->dma2usr_ratio * samples_to_bytes(runtime, pos) * 2;
		DPRINTK("copy channel=%x,pos=%x,frames=%x\n ",channel,pos,count);

	}
	else
	{
		hwbuf = runtime->dma_area + (channel * dma_csize) + samples_to_bytes(runtime, pos);
		DPRINTK("copy channel=%x,pos=%x,frames=%x\n ",channel,pos,count);
	}
	if(copy_from_user_with_conversion(state,hwbuf,(const char*)src,(tocount * stream->dma2usr_ratio)) <=0 )
	{
		printk(KERN_ERR "copy_from_user_with_conversion() failed\n");
		return -EFAULT;
	}

	return ret;
}

static int ep93xx_i2s_pcm_capture_copy(snd_pcm_substream_t * substream,int channel, 
				snd_pcm_uframes_t pos,void __user *src, snd_pcm_uframes_t count)
{
	audio_state_t *state 		= (audio_state_t *)(substream->private_data);
	snd_pcm_runtime_t *runtime 	= substream->runtime;
	audio_stream_t *stream 		= state->input_stream;
	int ret = 0;

	int tocount = frames_to_bytes(runtime,count);
	
	size_t dma_csize = runtime->dma_bytes / runtime->channels;
	char *hwbuf;

	if( channel == -1 )
	{
		hwbuf = runtime->dma_area  + stream->dma2usr_ratio * samples_to_bytes(runtime, pos) * 2;
		DPRINTK("copy channel=%x,pos=%x,frames=%x\n ",channel,pos,count);

	}
	else
	{
		hwbuf = runtime->dma_area + (channel * dma_csize) + samples_to_bytes(runtime, pos);
		DPRINTK("copy channel=%x,pos=%x,frames=%x\n ",channel,pos,count);
	}
	if(copy_to_user_with_conversion(state,(const char*)src,hwbuf,tocount) <= 0)
	{
		printk(KERN_ERR "copy_to_user_with_conversion() failed\n");
		return -EFAULT;
	}

	return ret;	
}

static snd_pcm_ops_t ep93xx_i2s_pcm_playback_ops = {
	.open		= ep93xx_i2s_pcm_open,
	.close		= ep93xx_i2s_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= ep93xx_i2s_pcm_hw_params,
	.hw_free	= ep93xx_i2s_pcm_hw_free,
	.prepare	= ep93xx_i2s_pcm_prepare,
	.trigger	= ep93xx_i2s_pcm_trigger,
	.pointer	= ep93xx_i2s_pcm_pointer,
	.copy		= ep93xx_i2s_pcm_playback_copy,
};

static snd_pcm_ops_t ep93xx_i2s_pcm_capture_ops = {
	.open		= ep93xx_i2s_pcm_open,
	.close		= ep93xx_i2s_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= ep93xx_i2s_pcm_hw_params,
	.hw_free	= ep93xx_i2s_pcm_hw_free,
	.prepare	= ep93xx_i2s_pcm_prepare,
	.trigger	= ep93xx_i2s_pcm_trigger,
	.pointer	= ep93xx_i2s_pcm_pointer,
	.copy 		= ep93xx_i2s_pcm_capture_copy,
};

static int ep93xx_i2s_pcm_new(snd_card_t *card, audio_state_t *state, snd_pcm_t **rpcm)
{
	snd_pcm_t *pcm;
	int play = state->output_stream? 1 : 0;/*SNDRV_PCM_STREAM_PLAYBACK*/
	int capt = state->input_stream ? 1 : 0;/*SNDRV_PCM_STREAM_CAPTURE*/
	int ret = 0;

	DPRINTK("ep93xx_i2s_pcm_new----ENTER\n");

	/* Register the new pcm device interface */
	ret = snd_pcm_new(card, "EP93xx-i2s-PCM", 0, play, capt, &pcm);
	if (ret){
		DPRINTK("%s--%x:card=%x,play=%x,capt=%x,&pcm=%x\n",__FUNCTION__,ret,card,play,capt,pcm);
		goto out;
	}

	/* allocate the pcm(DMA) memory */
	ret = snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,0, 128*1024, 128*1024);

	DPRINTK("The substream item : \n");
	DPRINTK("pcm->streams[0].substream->dma_buffer.addr  = 0x%x\n", pcm->streams[0].substream->dma_buffer.addr);
	DPRINTK("pcm->streams[0].substream->dma_buffer.area  = 0x%x\n", pcm->streams[0].substream->dma_buffer.area);
	DPRINTK("pcm->streams[0].substream->dma_buffer.bytes = 0x%x\n", pcm->streams[0].substream->dma_buffer.bytes);
	DPRINTK("pcm->streams[1].substream->dma_buffer.addr  = 0x%x\n", pcm->streams[1].substream->dma_buffer.addr);
	DPRINTK("pcm->streams[1].substream->dma_buffer.area  = 0x%x\n", pcm->streams[1].substream->dma_buffer.area);
	DPRINTK("pcm->streams[1].substream->dma_buffer.bytes = 0x%x\n", pcm->streams[1].substream->dma_buffer.bytes);	

	pcm->private_data = state;
	
	/* seem to free the pcm data struct-->self dma buffer */
	pcm->private_free = (void*) snd_pcm_lib_preallocate_free_for_all;

	/* alsa pcm ops setting for SNDRV_PCM_STREAM_PLAYBACK */
	if (play) {
		int stream = SNDRV_PCM_STREAM_PLAYBACK;
		snd_pcm_set_ops(pcm, stream, &ep93xx_i2s_pcm_playback_ops);
		
	}

	/* alsa pcm ops setting for SNDRV_PCM_STREAM_CAPTURE */	
	if (capt) {
		int stream = SNDRV_PCM_STREAM_CAPTURE;
		snd_pcm_set_ops(pcm, stream, &ep93xx_i2s_pcm_capture_ops);
	}

	if (rpcm)
		*rpcm = pcm;
	
	DPRINTK("ep93xx_i2s_pcm_new----EXIT\n");
 out:
	return ret;
}

static int ep93xx_i2s_volume_info(snd_kcontrol_t *kcontrol, 
					snd_ctl_elem_info_t * uinfo)
{
	//audio_state_t *state = (audio_state_t *)(kcontrol->private_data);
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 100;
	return 0;
}

static int ep93xx_i2s_volume_get(snd_kcontrol_t * kcontrol,
                                       snd_ctl_elem_value_t * ucontrol)
{
	//unsigned long flags;
	//audio_state_t *state = (audio_state_t *)(kcontrol->private_data);

	return 0;
}

static int ep93xx_i2s_volume_put(snd_kcontrol_t * kcontrol,
                                       snd_ctl_elem_value_t * ucontrol)
{
	//unsigned long flags;
	//audio_state_t *state = (audio_state_t *)(kcontrol->private_data);
	int val;
	
	//spin_lock_irqsave(&emu->reg_lock, flags);

	val = ucontrol->value.integer.value[0] & 255;
	ep93xx_set_volume( val );
	
	//spin_unlock_irqrestore(&emu->reg_lock, flags);
	return 1;
}

static snd_kcontrol_new_t ep93xx_i2s_volume_control =
{
	.access		= SNDRV_CTL_ELEM_ACCESS_READWRITE | SNDRV_CTL_ELEM_ACCESS_INACTIVE,
	.iface		= SNDRV_CTL_ELEM_IFACE_MIXER,
	.name		= "First Playback Volume",
	.count		= 1,
	.info		= ep93xx_i2s_volume_info,
	.get		= ep93xx_i2s_volume_get,
	.put		= ep93xx_i2s_volume_put,
};

static int __init ep93xx_i2s_mixer_new(snd_card_t *card, audio_state_t *state)
{
	int idx, err;

	snd_assert(card != NULL, return -EINVAL);

	for (idx = 0; idx < 1; idx++) {
		if ((err = snd_ctl_add(card, snd_ctl_new1(&ep93xx_i2s_volume_control, state))) < 0)
			return err;
	}

	strcpy(card->mixername, "EP93XX MIXER");
	return 0;
}

/* module power management */
#ifdef CONFIG_PM
static int ep93xx_i2s_suspend(struct device * dev, u32 state, u32 level)
{
	return 0;
}

static int ep93xx_i2s_resume(struct device * dev, u32 level)
{
	return 0;
}
#endif /* COMFIG_PM */

/* module init & exit */
static int ep93xx_i2s_probe(struct device *dev)
{
	snd_card_t *card;
	int err = -ENOMEM;

	/*
	 * Open an instance of the SSP driver for I2S codec register access.
	 */
	SSP_Handle = SSPDriver->Open(I2S_CODEC,0);
	if( SSP_Handle < 1 )
	{
		DPRINTK( "Couldn't open SSP driver!\n");
		return -ENODEV;
	}

	/*
	 * Enable audio early on, give the DAC time to come up.
	 */ 
	ep93xx_audio_init( 0 );

	/* register the soundcard */
	card = snd_card_new(-1, 0, THIS_MODULE, 0);

	card->dev = dev;

	if (card == NULL) {
		DPRINTK(KERN_ERR "snd_card_new() failed\n");
		goto error;
	}

	strcpy(card->driver,    "ep93xx i2s");
	strcpy(card->shortname, "ep93xx i2s audio");
	strcpy(card->longname,  "Cirrus Logic ep93xx i2s audio");

	/* pcm */
	if (ep93xx_i2s_pcm_new(card, &audio_state, &ep93xx_i2s_pcm) < 0)
		goto error;

	/* mixer */
	if (ep93xx_i2s_mixer_new(card, &audio_state) < 0)
		goto error;

	if ((err = snd_card_register(card)) == 0) {
		printk( KERN_INFO "Cirrus Logic ep93xx i2s audio initialized\n" );
		return 0;
	}

error:
	snd_card_free(card);
	return err;
}

static int ep93xx_i2s_remove(struct device *dev)
{
	snd_card_t *card = dev_get_drvdata(dev);

	if (card) {
		snd_card_free(card);
		dev_set_drvdata(dev, NULL);
	}

	return 0;
}

static struct device_driver ep93xx_i2s_device = {
	.name       = "ep93xxi2s",
	.bus        = &platform_bus_type,
	.probe      = ep93xx_i2s_probe,
	.remove     = ep93xx_i2s_remove,
#ifdef CONFIG_PM
	.suspend    = ep93xx_i2s_suspend,
	.resume     = ep93xx_i2s_resume,
#endif
};

static int __init ep93xx_i2s_init(void)
{
	return driver_register(&ep93xx_i2s_device);
}

static void __exit ep93xx_i2s_exit(void)
{
	driver_unregister(&ep93xx_i2s_device);
}

module_init(ep93xx_i2s_init);
module_exit(ep93xx_i2s_exit);

MODULE_AUTHOR("Fred Wei");
MODULE_DESCRIPTION("Cirrus Logic EP93xx I2S audio module");
MODULE_LICENSE("GPL");
